""" Fast Fragment Seeker
Extracts recurring tree fragments from constituency treebanks.

NB: there is a known bug in multiprocessing which makes it impossible to detect
Ctrl-C or fatal errors like segmentation faults in child processes which causes
the master program to wait forever for output from its children. Therefore if
you want to abort, kill the program manually (e.g., press Ctrl-Z and issue
'kill %1'). If the program seems stuck, re-run without multiprocessing
(pass --numproc 1) to see if there might be a bug. """

from __future__ import division, print_function
import io
import os
import re
import sys
import codecs
import logging
from multiprocessing import Pool, cpu_count, log_to_stderr, SUBDEBUG
from collections import defaultdict
from itertools import count
from getopt import gnu_getopt, GetoptError
from discodop.tree import Tree
from discodop.treetransforms import binarize, introducepreterminals
from discodop._fragments import extractfragments, fastextractfragments, \
		exactcounts, readtreebank, getctrees, completebitsets, coverbitsets

USAGE = """Usage: %s <treebank1> [treebank2] [options]
  or: %s --batch=<dir> <treebank1> <treebank2>... [options]
If only one treebank is given, fragments occurring at least twice are sought.
If two treebanks are given, finds common fragments between first & second.
Input is in Penn treebank format (S-expressions), one tree per line.
Output contains lines of the form "tree<TAB>frequency".
Frequencies refer to the last treebank by default.
Output is sent to stdout; to save the results, redirect to a file.
Options:
  --disc        work with discontinuous trees; input is in Negra export format.
                output: tree<TAB>sentence<TAB>frequency
                where "tree' has indices as leaves, referring to elements of
                "sentence", a space separated list of words.
  --indices     report sets of indices instead of frequencies.
  --cover       include 'cover' fragments corresponding to single productions.
  --complete    find complete matches of fragments from treebank1 (needle) in
                treebank2 (haystack).
  --batch=dir   enable batch mode; any number of treebanks > 1 can be given;
                first treebank will be compared to all others, frequencies
                refer to first treebank.
                Results are written to filenames of the form dir/A_B.
  --numproc=n   use n independent processes, to enable multi-core usage
                (default: 1); use 0 to detect the number of CPUs.
  --numtrees=n  only read first n trees from first treebank
  --encoding=x  use x as treebank encoding, e.g. UTF-8, ISO-8859-1, etc.
  --approx      report approximate frequencies (lower bound)
  --nofreq      do not report frequencies.
  --quadratic   use the slower, quadratic algorithm for finding fragments.
  --alt         alternative output format: (NP (DT "a") NN)
                default: (NP (DT a) (NN ))
  --debug       extra debug information, ignored when numproc > 1.
  --quiet       disable all messages.""" % (sys.argv[0], sys.argv[0])

FLAGS = ('approx', 'indices', 'nofreq', 'complete', 'complement',
		'disc', 'quiet', 'debug', 'quadratic', 'cover', 'alt')
OPTIONS = ('numproc=', 'numtrees=', 'encoding=', 'batch=')
PARAMS = {}
FRONTIERRE = re.compile(r"\(([^ ()]+) \)")
TERMRE = re.compile(r"\(([^ ()]+) ([^ ()]+)\)")
APPLY = lambda x, _y: x()


def main(argv=None):
	""" Command line interface to fragment extraction. """
	if argv is None:
		argv = sys.argv
	try:
		opts, args = gnu_getopt(argv[1:], '', FLAGS + OPTIONS)
	except GetoptError as err:
		print("%s\n%s" % (err, USAGE))
		return
	opts = dict(opts)

	for flag in FLAGS:
		PARAMS[flag] = '--' + flag in opts
	numproc = int(opts.get("--numproc", 1))
	if numproc == 0:
		numproc = cpu_count()
	limit = int(opts.get('--numtrees', 0))
	encoding = opts.get("--encoding", "UTF-8")
	batchdir = opts.get("--batch")

	if len(args) < 1:
		print("missing treebank argument")
	if batchdir is None and len(args) not in (1, 2):
		print("incorrect number of arguments:", args)
		print(USAGE)
		return
	if batchdir:
		assert numproc == 1, ("Batch mode only supported in single-process "
				"mode. Use the xargs command for multi-processing.")
	if args[0] == '-':
		args[0] = '/dev/stdin'
	for a in args:
		assert os.path.exists(a), "not found: %r" % a
	if PARAMS['complete']:
		assert len(args) == 2 or batchdir, (
				"need at least two treebanks with --complete.")
		assert not PARAMS['approx'] and not PARAMS['nofreq'], (
				"--complete is incompatible with --nofreq and --approx")

	level = logging.WARNING if PARAMS['quiet'] else logging.DEBUG
	logging.basicConfig(level=level, format='%(message)s')
	if PARAMS['debug'] and numproc > 1:
		logger = log_to_stderr()
		logger.setLevel(SUBDEBUG)

	logging.info("Fast Fragment Seeker")

	assert numproc
	logging.info("parameters:\n%s", "\n".join("    %s:\t%r" % kv
		for kv in sorted(PARAMS.items())))
	logging.info("\n".join("treebank%d: %s" % (n + 1, a)
		for n, a in enumerate(args)))

	if numproc == 1 and batchdir:
		batch(batchdir, args, limit, encoding)
	else:
		fragmentkeys, counts = regular(args, numproc, limit, encoding)
		printfragments(fragmentkeys, counts)


def regular(filenames, numproc, limit, encoding):
	""" non-batch processing. multiprocessing optional. """
	mult = 1
	if PARAMS['approx']:
		fragments = defaultdict(int)
	else:
		fragments = {}
	# detect corpus reading errors in this process (e.g., wrong encoding)
	initworker(filenames[0], filenames[1] if len(filenames) == 2 else None,
			limit, encoding)
	if numproc == 1:
		mymap = map
		myapply = APPLY
	else:  # multiprocessing, start worker processes
		pool = Pool(processes=numproc, initializer=initworker,
			initargs=(filenames[0],
				filenames[1] if len(filenames) == 2 else None, limit, encoding))
		mymap = pool.imap
		myapply = pool.apply
	numtrees = (PARAMS['trees1'].len if limit == 0
			else min(PARAMS['trees1'].len, limit))

	if PARAMS['complete']:
		fragments = completebitsets(PARAMS['trees1'], PARAMS['sents1'],
				PARAMS['labels'], PARAMS['disc'])
	else:
		if len(filenames) == 1:
			work = workload(numtrees, mult, numproc)
		else:
			chunk = numtrees // (mult * numproc) + 1
			work = [(a, a + chunk) for a in range(0, numtrees, chunk)]
		if numproc != 1:
			logging.info("work division:\n%s", "\n".join("    %s:\t%r" % kv
				for kv in sorted(dict(numchunks=len(work), mult=mult).items())))
		dowork = mymap(worker, work)
		for n, results in enumerate(dowork):
			if PARAMS['approx']:
				for frag, x in results.items():
					fragments[frag] += x
			else:
				fragments.update(results)
	if PARAMS['cover']:
		assert not PARAMS['trees2'], "only supported for single treebank."
		cover = myapply(coverfragworker, ()).get()
		if PARAMS['approx']:
			fragments.update(zip(cover,
					exactcounts(PARAMS['trees1'], PARAMS['trees1'],
					cover.values(), fast=not PARAMS['quadratic'])))
		else:
			fragments.update(cover)
		logging.info("merged %d cover fragments", len(cover))
	fragmentkeys = list(fragments)
	if PARAMS['nofreq']:
		counts = None
	elif PARAMS['approx']:
		counts = [fragments[a] for a in fragmentkeys]
	else:
		task = "indices" if PARAMS['indices'] else "counts"
		logging.info("dividing work for exact %s", task)
		bitsets = [fragments[a] for a in fragmentkeys]
		countchunk = len(bitsets) // numproc + 1
		work = list(range(0, len(bitsets), countchunk))
		work = [(n, len(work), bitsets[a:a + countchunk])
				for n, a in enumerate(work)]
		counts = []
		logging.info("getting exact %s", task)
		for a in mymap(exactcountworker, work):
			counts.extend(a)
	if numproc != 1:
		pool.close()
		pool.join()
		del dowork, pool
	return fragmentkeys, counts


def batch(outputdir, filenames, limit, encoding):
	""" batch processing: three or more treebanks specified.
	The use case for this is when you have one big treebank which you want to
	compare to lots of smaller sets of trees, and get the results for each
	comparison in a separate file. """
	initworker(filenames[0], None, limit, encoding)
	trees1 = PARAMS['trees1']
	sents1 = PARAMS['sents1']
	if PARAMS['approx']:
		fragments = defaultdict(int)
	else:
		fragments = {}
	for filename in filenames[1:]:
		PARAMS.update(read2ndtreebank(filename, PARAMS['labels'],
			PARAMS['prods'], PARAMS['disc'], limit, encoding))
		trees2 = PARAMS['trees2']
		sents2 = PARAMS['sents2']
		if PARAMS['complete']:
			fragments = completebitsets(PARAMS['trees2'], sents2,
					PARAMS['labels'], PARAMS['disc'])
		elif PARAMS['quadratic']:
			fragments = extractfragments(trees2, sents2, 0, 0,
					PARAMS['labels'], trees1, sents1,
					discontinuous=PARAMS['disc'], debug=PARAMS['debug'],
					approx=PARAMS['approx'])
		else:
			fragments = fastextractfragments(trees2, sents2, 0, 0,
					PARAMS['labels'], trees1, sents1,
					discontinuous=PARAMS['disc'], debug=PARAMS['debug'],
					approx=PARAMS['approx'])
		counts = None
		fragmentkeys = list(fragments)
		if PARAMS['approx'] or not fragments:
			counts = fragments.values()
		elif not PARAMS['nofreq']:
			bitsets = [fragments[a] for a in fragmentkeys]
			logging.info("getting %s for %d fragments",
					"indices of occurrence" if PARAMS['indices']
					else "exact counts", len(bitsets))
			counts = exactcounts(trees2, trees1, bitsets,
					indices=PARAMS['indices'], fast=not PARAMS['quadratic'])
		outputfilename = "%s/%s_%s" % (outputdir,
				os.path.basename(filenames[0]), os.path.basename(filename))
		out = io.open(outputfilename, "w", encoding=encoding)
		printfragments(fragmentkeys, counts, out=out)
		logging.info("wrote to %s", outputfilename)


def readtreebanks(treebank1, treebank2=None, discontinuous=False,
		limit=0, encoding="utf-8"):
	""" Read one or two treebanks.  """
	labels = []
	prods = {}
	trees1, sents1 = readtreebank(treebank1, labels, prods,
			not PARAMS['quadratic'], discontinuous, limit, encoding)
	trees2, sents2 = readtreebank(treebank2, labels, prods,
			not PARAMS['quadratic'], discontinuous, limit, encoding)
	trees1.indextrees(prods)
	if trees2:
		trees2.indextrees(prods)
	return dict(trees1=trees1, sents1=sents1, trees2=trees2, sents2=sents2,
		prods=prods, labels=labels)


def read2ndtreebank(treebank2, labels, prods, discontinuous=False,
	limit=0, encoding="utf-8"):
	""" Read a second treebank.  """
	trees2, sents2 = readtreebank(treebank2, labels, prods,
		not PARAMS['quadratic'], discontinuous, limit, encoding)
	logging.info("%r: %d trees; %d nodes (max %d). "
			"labels: %d, prods: %d",
			treebank2, len(trees2), trees2.numnodes, trees2.maxnodes,
			len(set(PARAMS['labels'])), len(PARAMS['prods']))
	return dict(trees2=trees2, sents2=sents2, prods=prods, labels=labels)


def initworker(treebank1, treebank2, limit, encoding):
	""" Read treebanks for this worker. We do this separately for each process
	under the assumption that this is advantageous with a NUMA architecture. """
	PARAMS.update(readtreebanks(treebank1, treebank2,
		limit=limit, discontinuous=PARAMS['disc'], encoding=encoding))
	if PARAMS['debug']:
		print("\nproductions:")
		for a, b in sorted(PARAMS['prods'].items(), key=lambda x: x[1]):
			print(b, a[0], '=>', ' '.join(a[1:]))
	trees1 = PARAMS['trees1']
	assert trees1
	m = "treebank1: %d trees; %d nodes (max: %d);" % (
		trees1.len, trees1.numnodes, trees1.maxnodes)
	if treebank2:
		trees2 = PARAMS['trees2']
		assert trees2
		m += " treebank2: %d trees; %d nodes (max %d);" % (
			trees2.len, trees2.numnodes, trees2.maxnodes)
	logging.info("%s labels: %d, prods: %d", m, len(set(PARAMS['labels'])),
		len(PARAMS['prods']))


def initworkersimple(trees, sents, trees2=None, sents2=None):
	""" A simpler initialization for a worker in which a treebank has already
	been loaded. """
	PARAMS.update(getctrees(trees, sents, trees2, sents2))
	assert PARAMS['trees1']


def worker(interval):
	""" Worker function that initiates the extraction of fragments
	in each process. """
	offset, end = interval
	trees1 = PARAMS['trees1']
	trees2 = PARAMS['trees2']
	sents1 = PARAMS['sents1']
	sents2 = PARAMS['sents2']
	assert offset < trees1.len
	result = {}
	if PARAMS['quadratic']:
		result = extractfragments(trees1, sents1, offset, end,
				PARAMS['labels'], trees2, sents2, approx=PARAMS['approx'],
				discontinuous=PARAMS['disc'], debug=PARAMS['debug'])
	else:
		result = fastextractfragments(trees1, sents1, offset, end,
				PARAMS['labels'], trees2, sents2, approx=PARAMS['approx'],
				discontinuous=PARAMS['disc'], complement=PARAMS['complement'],
				debug=PARAMS.get('debug'))
	logging.debug("finished %d--%d", offset, end)
	return result


def exactcountworker(args):
	""" Worker function that initiates the counting of fragments
	in each process. """
	n, m, bitsets = args
	trees1 = PARAMS['trees1']
	results = exactcounts(trees1, PARAMS['trees2'] or trees1, bitsets,
			fast=not PARAMS['quadratic'], indices=PARAMS['indices'])
	if PARAMS['complete']:
		logging.debug("complete matches %d of %d", n + 1, m)
	elif PARAMS['indices']:
		logging.debug("exact indices %d of %d", n + 1, m)
	else:
		logging.debug("exact counts %d of %d", n + 1, m)
	return results


def coverfragworker():
	""" Worker function that gets depth-1 fragments. Does not need
	multiprocessing but using it avoids reading the treebank again. """
	trees1 = PARAMS['trees1']
	trees2 = PARAMS['trees2']
	return coverbitsets(trees1, PARAMS['sents1'], PARAMS['labels'],
			max(trees1.maxnodes, (trees2 or trees1).maxnodes),
			PARAMS['disc'])


def workload(numtrees, mult, numproc):
	""" Get an even workload. When n trees are compared against themselves,
	n * (n - 1) total comparisons are made.
	Each tree m has to be compared to all trees x such that m < x < n.
	(meaning there are more comparisons for lower n).

	:returns: a sequence of (start, end) intervals such that
		the number of comparisons is approximately balanced. """
	# could base on number of nodes as well.
	if numproc == 1:
		return [(0, numtrees)]
	# here chunk is the number of tree pairs that will be compared
	goal = togo = total = 0.5 * numtrees * (numtrees - 1)
	chunk = total // (mult * numproc) + 1
	goal -= chunk
	result = []
	last = 0
	for n in range(1, numtrees):
		togo -= numtrees - n
		if togo <= goal:
			goal -= chunk
			result.append((last, n))
			last = n
	if last < numtrees:
		result.append((last, numtrees))
	return result


def getfragments(trees, sents, numproc=1, iterate=False, complement=False):
	""" Get recurring fragments with exact counts in a single treebank.

	:returns: a dictionary whose keys are fragments as strings, and
		frequencies / indices as values.
	:param trees:  a sequence of binarized Tree objects. """
	if numproc == 0:
		numproc = cpu_count()
	numtrees = len(trees)
	assert numtrees
	mult = 1  # 3 if numproc > 1 else 1
	fragments = {}
	trees = trees[:]
	work = workload(numtrees, mult, numproc)
	PARAMS.update(disc=True, indices=True, approx=False, complete=False,
			quadratic=False, complement=complement)
	if numproc == 1:
		initworkersimple(trees, list(sents))
		mymap = map
		myapply = APPLY
	else:
		logging.info("work division:\n%s", "\n".join("    %s: %r" % kv
			for kv in sorted(dict(numchunks=len(work),
				numproc=numproc).items())))
		# start worker processes
		pool = Pool(processes=numproc, initializer=initworkersimple,
			initargs=(trees, list(sents)))
		mymap = pool.map
		myapply = pool.apply
	# collect recurring fragments
	logging.info("extracting recurring fragments")
	for a in mymap(worker, work):
		fragments.update(a)
	# add 'cover' fragments corresponding to single productions
	cover = myapply(coverfragworker, ())
	before = len(fragments)
	fragments.update(cover)
	logging.info("merged %d unseen cover fragments", len(fragments) - before)
	fragmentkeys = list(fragments)
	bitsets = [fragments[a] for a in fragmentkeys]
	countchunk = len(bitsets) // numproc + 1
	work = list(range(0, len(bitsets), countchunk))
	work = [(n, len(work), bitsets[a:a + countchunk])
			for n, a in enumerate(work)]
	logging.info("getting exact counts for %d fragments", len(bitsets))
	counts = []
	for a in mymap(exactcountworker, work):
		counts.extend(a)
	if numproc != 1:
		pool.close()
		pool.join()
		del pool
	if iterate:  # optionally collect fragments of fragments
		logging.info("extracting fragments of recurring fragments")
		PARAMS['complement'] = False  # needs to be turned off if it was on
		newfrags = fragments
		trees, sents = None, None
		ids = count()
		for _ in range(10):  # up to 10 iterations
			newtrees = [binarize(
					introducepreterminals(Tree.parse(tree, parse_leaf=int),
					ids=ids), childchar="}") for tree, _ in newfrags]
			newsents = [["#%d" % next(ids) if word is None else word
					for word in sent] for _, sent in newfrags]
			newfrags, newcounts = iteratefragments(
					fragments, newtrees, newsents, trees, sents, numproc)
			if len(newfrags) == 0:
				break
			if trees is None:
				trees = []
				sents = []
			trees.extend(newtrees)
			sents.extend(newsents)
			fragmentkeys.extend(newfrags)
			counts.extend(newcounts)
			fragments.update(zip(newfrags, newcounts))
	logging.info("found %d fragments", len(fragmentkeys))
	return dict(zip(fragmentkeys, counts))


def iteratefragments(fragments, newtrees, newsents, trees, sents, numproc):
	""" Get fragments of fragments. """
	numtrees = len(newtrees)
	assert numtrees
	if numproc == 1:  # set fragments as input
		initworkersimple(newtrees, newsents, trees, sents)
		mymap = map
	else:
		# since the input trees change, we need a new pool each time
		pool = Pool(processes=numproc, initializer=initworkersimple,
			initargs=(newtrees, newsents, trees, sents))
		mymap = pool.imap
	newfragments = {}
	for a in mymap(worker, workload(numtrees, 1, numproc)):
		newfragments.update(a)
	logging.info("before: %d, after: %d, difference: %d",
		len(fragments), len(set(fragments) | set(newfragments)),
		len(set(newfragments) - set(fragments)))
	# we have to get counts for these separately because they're not coming
	# from the same set of trees
	newkeys = list(set(newfragments) - set(fragments))
	bitsets = [newfragments[a] for a in newkeys]
	countchunk = len(bitsets) // numproc + 1
	if countchunk == 0:
		return newkeys, []
	work = list(range(0, len(bitsets), countchunk))
	work = [(n, len(work), bitsets[a:a + countchunk])
			for n, a in enumerate(work)]
	logging.info("getting exact counts for %d fragments", len(bitsets))
	counts = []
	for a in mymap(exactcountworker, work):
		counts.extend(a)
	if numproc != 1:
		pool.close()
		pool.join()
		del pool
	return newkeys, counts


def altrepr(a):
	""" Alternative format
	Replace double quotes with double single quotes: " -> ''
	Quote terminals with double quotes terminal: -> "terminal"
	Remove parentheses around frontier nodes: (NN ) -> NN

	>>> altrepr("(NP (DT a) (NN ))")
	'(NP (DT "a") NN)'
	"""
	return FRONTIERRE.sub(r'\1', TERMRE.sub(r'(\1 "\2")', a.replace('"', "''")))


def printfragments(fragments, counts, out=None):
	""" Dump fragments to standard output or some other file object. """
	if out is None:
		out = sys.stdout
		if sys.stdout.encoding is None:
			out = codecs.getwriter('utf-8')(out)
	if PARAMS['complete']:
		logging.info("total number of matches: %d", sum(counts))
	else:
		logging.info("number of fragments: %d", len(fragments))
	if PARAMS['nofreq']:
		for a in fragments:
			if PARAMS['alt']:
				a = altrepr(a)
			out.write("%s\n" % (("%s\t%s" % (a[0],
					' '.join("%s" % x if x else '' for x in a[1])))
					if PARAMS['disc'] else a.decode('utf-8')))
		return
	# when comparing two treebanks, a frequency of 1 is normal;
	# otherwise, raise alarm.
	if PARAMS.get('trees2') or PARAMS['cover'] or PARAMS['complement']:
		threshold = 0
	else:
		threshold = 1
	if PARAMS['indices']:
		for a, theindices in zip(fragments, counts):
			if PARAMS['alt']:
				a = altrepr(a)
			if len(theindices) > threshold:
				out.write("%s\t%r\n" % (("%s\t%s" % (a[0],
					' '.join("%s" % x if x else '' for x in a[1])))
					if PARAMS['disc'] else a.decode('utf-8'),
					list(sorted(theindices.elements()))))
			elif threshold:
				raise ValueError("invalid fragment--frequency=1: %r" % a)
				#logging.warning("invalid fragment--frequency=1: %r", a)
		return
	for a, freq in zip(fragments, counts):
		if freq > threshold:
			if PARAMS['alt']:
				a = altrepr(a)
			out.write("%s\t%d\n" % (("%s\t%s" % (a[0],
				' '.join("%s" % x if x else '' for x in a[1])))
				if PARAMS['disc'] else a, freq))
		elif threshold:
			raise ValueError("invalid fragment--frequency=1: %r" % a)
			#logging.warning("invalid fragment--frequency=1: %r", a)


def test():
	""" Simple test. """
	main("fragments.py --disc alpinosample.export".split())

if __name__ == '__main__':
	main()



from libc.stdlib cimport malloc, calloc, free, qsort
cimport cython
from discodop.containers cimport ULLong, ULong, UInt, UChar, Rule, LexicalRule

cdef extern from "macros.h":
	int BITNSLOTS(int nb)
	void SETBIT(ULong a[], int b)


@cython.final
cdef class Grammar:
	cdef Rule **unary, **lbinary, **rbinary, **bylhs
	cdef ULong *chainvec
	cdef UInt *mapping, **splitmapping
	cdef UChar *fanout
	cdef readonly currentmodel
	cdef readonly size_t nonterminals, numrules, numunary, numbinary, maxfanout
	cdef readonly bint logprob, bitpar
	cdef readonly object models
	cdef readonly bytes origrules, start
	cdef readonly unicode origlexicon
	cdef readonly list tolabel, lexical, modelnames, rulemapping
	cdef readonly dict toid, lexicalbyword, lexicalbylhs, lexicalbynum, rulenos
	cdef _convertrules(Grammar self, list rulelines, dict fanoutdict)
	cdef _indexrules(Grammar self, Rule **dest, int idx, int filterlen)
	cdef rulestr(self, Rule rule)
	cdef yfstr(self, Rule rule)



from discodop.containers cimport Edge, ChartItem, LCFRSEdge
from cpython cimport PyList_GET_ITEM, PyList_GET_SIZE
cimport cython

@cython.final
cdef class Entry:
	cdef object key
	cdef object value
	cdef unsigned long count
ctypedef bint (*CmpFun)(Entry a, Entry b)

cdef inline Entry new_Entry(object k, object v, unsigned long c):
	cdef Entry entry = Entry.__new__(Entry)
	entry.key = k
	entry.value = v
	entry.count = c
	return entry

@cython.final
cdef class Agenda:
	cdef unsigned long length, counter
	cdef list heap
	cdef dict mapping
	cdef setitem(self, key, object value)
	cdef setifbetter(self, key, object value)
	cdef object getitem(self, key)
	cdef object replace(self, object key, object value)
	cdef Entry popentry(self)
	cdef Entry peekentry(self)
	cdef bint contains(self, key)

@cython.final
cdef class EdgeAgenda:
	cdef unsigned long length, counter
	cdef list heap
	cdef dict mapping
	cdef setitem(self, key, object value)
	cdef setifbetter(self, key, object value)
	cdef LCFRSEdge getitem(self, key)
	cdef LCFRSEdge replace(self, object key, object value)
	cdef Entry popentry(self)
	cdef Entry peekentry(self)
	cdef bint contains(self, key)

cdef list nsmallest(int n, object iterable)



""" NB: most functions are in bit.pxd to facilitate function inlining. """


def pyintbitcount(a):
	""" Number of set bits (1s)
	>>> bitcount(0b0011101)
	4
	"""
	count = 0
	while a:
		a &= a - 1
		count += 1
	return count


cpdef bint testbitshort(unsigned short arg, UInt pos):
	""" Mask a particular bit, return nonzero if set
	>>> testbitshort(0b0011101, 0)
	True
	>>> testbitshort(0b0011101, 1)
	False """
	return (arg >> pos) & 1


cpdef bint testbitc(UChar arg, UInt pos):
	""" Mask a particular bit, return nonzero if set
	>>> testbitc(0b0011101, 0)
	True
	>>> testbitc(0b0011101, 1)
	False """
	return (arg >> pos) & 1


cpdef int bitcount(ULLong vec):
	""" Number of set bits (1s)
	>>> bitcount(0b0011101)
	4
	"""
	return __builtin_popcountll(vec)


cpdef int pyintnextset(a, int pos):
	""" First set bit, starting from pos
	>>> pyintnextset(0b001101, 1)
	2 """
	cdef ULong mask = -1
	a >>= pos
	if a == 0:
		return -1
	while a & mask == 0:
		a >>= (8 * sizeof(ULong))
		pos += (8 * sizeof(ULong))
	return pos + __builtin_ctzl(a & mask)


cpdef bint bitminmax(ULLong a, ULLong b):
	""" test whether the leftmost bit of b is adjacent to the first component
	of a. """
	return nextset(b, 0) == nextunset(a, nextset(a, 0))


cpdef int fanout(arg):
	""" number of contiguous components in bit vector (gaps plus one)
	>>> fanout(0b011011011)
	3
	"""
	cdef UInt result = 0
	cdef ULLong vec
	if arg < ((1ULL << (8ULL * sizeof(ULLong) - 1ULL)) - 1ULL):
		vec = arg
		while vec:
			vec >>= __builtin_ctzll(vec)
			vec >>= __builtin_ctzll(~vec)
			result += 1
	else:
		# when the argument does not fit in unsigned long long, resort to plain
		# Python code to use Python's bigints. This algorithm is based
		# on a bit counting algorihm. It goes through as many iterations as
		# there are set bits (while the other algorithm jumps over contiguous
		# sequences of set bits in one go).
		prev = arg
		result = 0
		while arg:
			arg &= arg - 1
			if ((prev - arg) << 1) & prev == 0:
				result += 1
			prev = arg
	return result


def main():
	cdef ULong ulongvec[2]
	bigpyint = 0b11100110101111001101011111100110101001100110
	assert nextset(0b001100110, 3) == 5
	assert nextset(0b001100110, 7) == -1
	assert nextunset(0b001100110, 1) == 3
	assert nextunset(~0ULL, 0) == 64
	assert pyintnextset(0b001100110, 3) == 5
	assert pyintnextset(0b001100110, 7) == -1
	assert pyintnextset(bigpyint, 3) == 5
	assert pyintnextset(bigpyint, 7) == 9
	assert bitcount(0b001100110) == 4
	assert bitlength(0b001100110) == 7
	assert bitminmax(0b001100, 0b110000)
	assert testbit(0b001100110, 1)
	assert not testbit(0b001100110, 3)
	assert fanout(0b0111100) == 1
	assert fanout(0b1000001) == 2
	assert fanout(0b011011011) == 3
	ulongvec[0] = 1UL << (sizeof(ULong) * 8 - 1)
	ulongvec[1] = 1
	assert anextset(ulongvec, 0, 2) == sizeof(ULong) * 8 - 1, (
			anextset(ulongvec, 0, 2), sizeof(ULong) * 8 - 1)
	assert anextset(ulongvec, sizeof(ULong) * 8, 2) == sizeof(ULong) * 8, (
		anextset(ulongvec, sizeof(ULong) * 8, 2), sizeof(ULong) * 8)
	assert anextunset(ulongvec, 0, 2) == 0, (
		anextunset(ulongvec, 0, 2), 0)
	assert anextunset(ulongvec, sizeof(ULong) * 8 - 1, 2) == (
			sizeof(ULong) * 8 + 1), (anextunset(ulongvec,
			sizeof(ULong) * 8 - 1, 2), sizeof(ULong) * 8 + 1)
	ulongvec[0] = 0
	assert anextset(ulongvec, 0, 2) == sizeof(ULong) * 8, (
		anextset(ulongvec, 0, 2), sizeof(ULong) * 8)
	ulongvec[1] = 0
	assert anextset(ulongvec, 0, 2) == -1, (
		anextset(ulongvec, 0, 2), -1)
	ulongvec[0] = ~0UL
	assert anextunset(ulongvec, 0, 2) == sizeof(ULong) * 8, (
		anextunset(ulongvec, 0, 2), sizeof(ULong) * 8)
	print('it worked')

if __name__ == '__main__':
	main()



""" Examples of various formalisms encoded in LCFRS grammars. """
from __future__ import print_function
from math import exp
from discodop import treetransforms, plcfrs, kbest
from discodop.tree import Tree
from discodop.grammar import treebankgrammar
from discodop._grammar import Grammar


def tree_adjoining_grammar():
	""" Example of a tree-adjoining grammar (TAG) encoded as an LCFRS.
	Taken from: Chen & Vijay-Shanker (2000), Automated extraction of TAGs
	from the Penn treebank.
	Limitations:
	- no epsilon productions
	Non-terminals have identifiers to encode elementary trees of depth > 2. """
	print("Tree-Adjoining Grammars in LCFRS")
	print("""initial trees:
(S (NP ) (VP (V fell)))
(NP (NN prices))
auxiliary trees:
(S (ADVP (RB Later) (S* ))
(VP (ADVP (RB drastically)) (VP* ))""")
	grammar = Grammar([
			((('ADVP#1', 'RB#1'), ((0, ), )), 1),
			((('ADVP#2', 'RB#2'), ((0, ), )), 1),
			((('NP', 'NN#1'), ((0, ), )), 1),
			((('ROOT', 'S'), ((0, ), )), 1),
			((('S', 'NP', 'VP'), ((0, 1), )), 1),
			((('S', 'ADVP#1', 'S'), ((0, 1), )), 1),
			((('VP', 'V#1'), ((0, ), )), 1),
			((('VP', 'ADVP#2', 'VP'), ((0, 1), )), 1),
			((('RB#1', 'Epsilon'), ('Later', )), 1),
			((('NN#1', 'Epsilon'), ('prices', )), 1),
			((('V#1', 'Epsilon'), ('fell', )), 1),
			((('RB#2', 'Epsilon'), ('drastically', )), 1)])
	print(grammar)
	assert parse(grammar, "prices fell".split())
	assert parse(grammar, "prices drastically fell".split())
	assert parse(grammar, "Later prices fell".split())
	assert parse(grammar, "Later prices drastically fell".split())

	# taken from: slides for course Grammar Formalisms, Kallmeyer (2011),
	# Mildly Context-Sensitive Grammar Formalisms:
	# LCFRS: Relations to other Formalisms
	print("the language {d} + {a**n b**m c**m d **n} with n>0, m>=0")
	print("""initial trees:
(S a (S Epsilon) F)
(F d)
auxiliary trees:
(S b S* c)""")
	grammar = Grammar([
			((('ROOT', 'a1'), ((0, ), )), 1),
			((('ROOT', 'a2'), ((0, ), )), 1),
			((('a1', 'a_b', 'a2'), ((0, 1), )), 1),
			((('a1', '_a', 'a2'), ((0, 1), )), 1),
			((('a2', '_d'), ((0, ), )), 1),
			((('a_b', '_a', 'b'), ((0, 1), )), 1),
			((('b', '_b', '_c'), ((0, 1), )), 1),
			((('b', 'b_2', 'b'), ((0, 1, 0), )), 1),
			((('b_2', '_b', '_c'), ((0, ), (1, ))), 1),
			((('_a', 'Epsilon'), ('a', )), 1),
			((('_b', 'Epsilon'), ('b', )), 1),
			((('_c', 'Epsilon'), ('c', )), 1),
			((('_d', 'Epsilon'), ('d', )), 1)])
	print(grammar)
	assert parse(grammar, list("d"))
	assert parse(grammar, list("ad"))
	assert parse(grammar, list("abcd"))
	assert parse(grammar, list("abbccd"))
	print("wrong:")
	assert not parse(grammar, list("abbbccd"))

	# Taken from: Boullier (1998), Generalization of Mildly
	# Context-Sensitive Formalisms.
	# Epsilon replaced with '|', added preterminal rules w/underscores
	print("the language { ww | w in {a,b}* }")
	print("""initial trees:
		(S (A Epsilon))
		auxiliary trees:
		(A a (A A*) a)
		(A b (A A*) b)
		(A (A A*))""")
	grammar = Grammar([
			((('_aa', '_a', '_a'), ((0, ), (1, ))), 1),
			((('_bb', '_b', '_b'), ((0, ), (1, ))), 1),
			((('A', '_aa', 'A'), ((0, 1), (0, 1))), 1),
			((('A', '_bb', 'A'), ((0, 1), (0, 1))), 1),
			((('A', '_aa'), ((0, ), (0, ))), 1),
			((('A', '_bb'), ((0, ), (0, ))), 1),
			((('ROOT', '_|'), ((0, ), )), 1),
			((('ROOT', 'A', '_|'), ((0, 1, 0), )), 1),
			((('_a', 'Epsilon'), ('a', )), 1),
			((('_b', 'Epsilon'), ('b', )), 1),
			((('_|', 'Epsilon'), ('|', )), 1)])
	print(grammar)
	assert parse(grammar, list("a|a"))
	assert parse(grammar, list("ab|ab"))
	assert parse(grammar, list("abaab|abaab"))
	print("wrong:")
	assert not parse(grammar, list("a|b"))
	assert not parse(grammar, list("aa|bb"))


def dependencygrammar():
	""" An example dependency structure encoded in an LCFRS grammar.
	Taken from: Gildea (2011), Optimal Parsing Strategies for Linear
	Context-Free Rewriting Systems.
	Limitations:
	- rules have to be binarized
	- lexical rules have to be unary
	These have been dealt with by introducing nodes w/underscores. """
	print("A dependency grammar in an LCFRS:")
	grammar = Grammar([
			((('NMOD', '_A'), ((0, ), )), 1),
			((('NMOD', '_the'), ((0, ), )), 1),
			((('NMOD_hearing', 'NMOD', '_hearing'), ((0, 1), )), 1),
			((('NP', 'NMOD', '_issue'), ((0, 1), )), 1),
			((('PP', '_on', 'NP'), ((0, 1, ), )), 1),
			((('ROOT', 'SBJ', 'is_VC'), ((0, 1, 0, 1), )), 1),
			((('SBJ', 'NMOD_hearing', 'PP'), ((0, ), (1, ))), 1),
			((('TMP', '_today'), ((0, ), )), 1),
			((('VC', '_scheduled', 'TMP'), ((0, ), (1, ))), 1),
			((('is_VC', '_is', 'VC'), ((0, 1), (1, ))), 1),
			((('_A', 'Epsilon'), ('A', )), 1),
			((('_hearing', 'Epsilon'), ('hearing', )), 1),
			((('_is', 'Epsilon'), ('is', )), 1),
			((('_scheduled', 'Epsilon'), ('scheduled', )), 1),
			((('_on', 'Epsilon'), ('on', )), 1),
			((('_the', 'Epsilon'), ('the', )), 1),
			((('_issue', 'Epsilon'), ('issue', )), 1),
			((('_today', 'Epsilon'), ('today', )), 1)])
	print(grammar)
	testsent = "A hearing is scheduled on the issue today".split()
	assert parse(grammar, testsent)


def bitext():
	""" Bitext parsing with a synchronous CFG.
	Translation would require a special decoder (instead of normal kbest
	derivations where the whole sentence is given). """
	print("bitext parsing with a synchronous CFG")
	trees = [Tree.parse(a, parse_leaf=int) for a in """\
	(ROOT (S (NP (NNP (John 0) (John 7))) (VP (VB (misses 1) (manque 5))\
		(PP (IN (a` 6)) (NP (NNP (Mary 2) (Mary 4)))))) (SEP (| 3)))
	(ROOT (S (NP (NNP (Mary 0) (Mary 4))) (VP (VB (likes 1) (aimes 5))\
		(NP (DT (la 6)) (NN (pizza 2) (pizza 7))))) (SEP (| 3)))""".split('\n')]
	sents = [["0"] * len(a.leaves()) for a in trees]
	for a in trees:
		treetransforms.binarize(a)
	compiled_scfg = Grammar(treebankgrammar(trees, sents))
	print("sentences:")
	for tree in trees:
		print(' '.join(w for _, w in sorted(tree.pos())))
	print("treebank:")
	for tree in trees:
		print(tree)
	print(compiled_scfg, "\n")

	print("correct translations:")
	assert parse(compiled_scfg, ["0"] * 7,
			"John likes Mary | John aimes Mary".split())
	assert parse(compiled_scfg, ["0"] * 9,
			"John misses pizza | la pizza manque a` John".split())

	print("incorrect translations:")
	assert not parse(compiled_scfg, ["0"] * 7,
			"John likes Mary | Mary aimes John".split())
	assert not parse(compiled_scfg, ["0"] * 9,
			"John misses pizza | John manque a` la pizza".split())

	# the following SCFG is taken from:
	# http://cdec-decoder.org/index.php?title=SCFG_translation
	# the grammar has been binarized and some new non-terminals had to be
	# introduced because terminals cannot appear in binary rules.
	lexicon = ("|", "ein", "ich", "Haus", "kleines", "grosses", "sah", "fand",
		"small", "little", "big", "large", "house", "shell", "a", "I",
		"saw", "found")
	another_scfg = Grammar([
			((('DT', '_ein', '_a'), ((0, ), (1, ))), 0.5),
			((('JJ', '_kleines', '_small'), ((0, ), (1, ))), 0.1),
			((('JJ', '_kleines', '_little'), ((0, ), (1, ))), 0.9),
			((('JJ', '_grosses', '_big'), ((0, ), (1, ))), 0.8),
			((('JJ', '_grosses', '_large'), ((0, ), (1, ))), 0.2345),
			((('NN_house', '_Haus', '_house'), ((0, ), (1, ))), 1),
			((('NN_shell', '_Haus', '_shell'), ((0, ), (1, ))), 1),
			((('NP', '_ich', '_I'), ((0, ), (1, ), )), 0.6),
			((('NP', 'DT', 'NP|<JJ-NN>'), ((0, 1), (0, 1))), 0.5),
			((('NP|<JJ-NN>', 'JJ', 'NN_house'), ((0, 1), (0, 1))), 0.1),
			((('NP|<JJ-NN>', 'JJ', 'NN_shell'), ((0, 1), (0, 1))), 1.3),
			((('ROOT', 'S', '_|'), ((0, 1, 0), )), 1),
			((('S', 'NP', 'VP'), ((0, 1), (0, 1))), 0.2),
			((('VP', 'V', 'NP'), ((0, 1), (0, 1))), 0.1),
			((('V', '_sah', '_saw'), ((0, ), (1, ))), 0.4),
			((('V', '_fand', '_found'), ((0, ), (1, ))), 0.4)]
			+ [((('_%s' % word, 'Epsilon'), (word, )), 1)
					for word in lexicon])
	print(another_scfg)
	sents = [
		"ich sah ein kleines Haus | I saw a small house".split(),
		"ich sah ein kleines Haus | I saw a little house".split(),
		"ich sah ein kleines Haus | I saw a small shell".split(),
		"ich sah ein kleines Haus | I saw a little shell".split()]
	for sent in sents:
		assert parse(another_scfg, sent), sent


def parse(compiledgrammar, testsent, testtags=None):
	""" Parse a sentence with a grammar. """
	chart, start, _ = plcfrs.parse(testsent,
		compiledgrammar, tags=testtags, exhaustive=True)
	print("input:", ' '.join("%d:%s" % a
			for a in enumerate(testtags if testtags else testsent)), end=' ')
	if start:
		print()
		results = kbest.lazykbest(chart, start, 10, compiledgrammar.tolabel)[0]
		for tree, prob in results:
			tree = Tree(tree)
			treetransforms.unbinarize(tree)
			print(exp(-prob), tree)
		print()
		return True
	else:
		print("no parse!\n")
		#parser.pprint_chart(chart, testsent, compiledgrammar.tolabel)
		return False


def main():
	""" Run all examples. """
	bitext()
	dependencygrammar()
	tree_adjoining_grammar()
	bitext()

if __name__ == '__main__':
	main()



""" Evaluation of (discontinuous) parse trees, following EVALB as much as
possible, as well as some alternative evaluation metrics.  """
from __future__ import division, print_function
import io
import sys
from getopt import gnu_getopt, GetoptError
from itertools import count
from collections import defaultdict, Counter as multiset
if sys.version[0] >= '3':
	from itertools import zip_longest  # pylint: disable=E0611
else:
	from itertools import izip_longest as zip_longest

from discodop.tree import Tree
from discodop.treedraw import DrawTree
from discodop.treebank import getreader, readheadrules, dependencies, splitpath
try:
	from discodop.treedist import treedist, newtreedist
except ImportError:
	from discodop.treedist import newtreedist as newtreedist, newtreedist

USAGE = 'Usage: %s <gold> <parses> [param] [options]' % sys.argv[0]
HELP = """\
Evaluation of (discontinuous) parse trees, following EVALB as much as possible.

%s
where gold and parses are files with parse trees, param is an EVALB parameter
file, and options may consist of:

  --cutofflen=n    Overrides the sentence length cutoff of the parameter file.
  --verbose        Print table with per sentence information.
  --debug          Print debug information with per sentence bracketings etc.
  --disconly       Only evaluate bracketings of discontinuous constituents
                   (only affects Parseval measures).
  --goldfmt        ...
  --parsesfmt      Specify corpus format. Options: export, bracket, discbracket
  --goldenc=enc    ...
  --parsesenc=enc  Specify a different encoding than the default UTF-8.
  --ted            Enable tree-edit distance evaluation.
  --headrules=x    Specify rules for head assignment of constituents that do
                   not already have a child marked as head; this enables
                   dependency evaluation.
  --functions=x    'remove'=default: strip functions off labels,
                   'leave': leave syntactic labels as is,
                   'add': evaluate both syntactic categories and functions,
                   'replace': only evaluate grammatical functions.
  --morphology=x   'no'=default: only evaluate POS tags,
                   'add': concatenate morphology tags to POS tags,
                   'replace': replace POS tags with morphology tags,
                   'between': add morphological node between POS tag and word.

The parameter file should be encoded in UTF-8 and supports the following
options (in addition to those described in the README of EVALB):
  DISC_ONLY        only consider discontinuous constituents for F-scores.
  TED              when enabled, give tree-edit distance scores; disabled by
                   default as these are slow to compute.
  DEBUG            -1 only print summary table
                   0 additionally, print category / tag breakdowns (default)
                     (after application of cutoff length).
                   1 give per-sentence results ('--verbose')
                   2 give detailed information for each sentence ('--debug')
  MAX_ERROR        this values is ignored, no errors are tolerated.
                   the parameter is accepted to support usage of unmodified
                   EVALB parameter files. """ % USAGE

HEADER = """
   Sentence                 Matched   Brackets            Corr      Tag
  ID Length  Recall  Precis Bracket   gold   test  Words  Tags Accuracy    LA\
""".splitlines()


def main():
	""" Command line interface for evaluation. """
	flags = ('test', 'verbose', 'debug', 'disconly', 'ted')
	options = ('goldenc=', 'parsesenc=', 'goldfmt=', 'parsesfmt=',
			'cutofflen=', 'headrules=', 'functions=', 'morphology=')
	try:
		opts, args = gnu_getopt(sys.argv[1:], '', flags + options)
	except GetoptError as err:
		print('error: %s\n%s' % (err, HELP))
		sys.exit(2)
	else:
		opts = dict(opts)
		if '--test' in opts:
			test()
			return
	try:
		assert 2 <= len(args) <= 3, 'Wrong number of arguments.\n%s' % (
				USAGE if len(args) else HELP)
		goldfile, parsesfile = args[:2]
		param = readparam(args[2] if len(args) == 3 else None)
		param['CUTOFF_LEN'] = int(opts.get('--cutofflen', param['CUTOFF_LEN']))
		param['DISC_ONLY'] = '--disconly' in opts
		param['DEBUG'] = max(param['DEBUG'],
				'--verbose' in opts, 2 * ('--debug' in opts))
		param['TED'] |= '--ted' in opts
		param['DEP'] = '--headrules' in opts
		if '--headrules' in opts:
			param['HEADRULES'] = readheadrules(opts['--headrules'])
	except AssertionError as err:
		print('error: %s' % err)
		sys.exit(2)
	goldreader = getreader(opts.get('--goldfmt', 'export'))
	parsesreader = getreader(opts.get('--parsesfmt', 'export'))
	gold = goldreader(*splitpath(goldfile),
			encoding=opts.get('--goldenc', 'utf-8'),
			functions=opts.get('--functions', 'remove'),
			morphology=opts.get('--morphology'))
	parses = parsesreader(*splitpath(parsesfile),
			encoding=opts.get('--parsesenc', 'utf-8'),
			functions=opts.get('--functions', 'remove'),
			morphology=opts.get('--morphology'))
	print(doeval(gold.parsed_sents(),
			gold.tagged_sents(),
			parses.parsed_sents(),
			parses.tagged_sents(),
			param))


def doeval(gold_trees, gold_sents, cand_trees, cand_sents, param):
	""" Do the actual evaluation on given parse trees and parameters.
	Results are printed to standard output. """
	assert gold_trees, 'no trees in gold file'
	assert cand_trees, 'no trees in parses file'
	keylen = max(len(str(x)) for x in cand_trees)
	if param['DEBUG'] == 1:
		print('Parameters:')
		for a in param:
			print('%s\t%s' % (a, param[a]))
		for a in HEADER:
			print(' ' * (keylen - 4) + a)
		print('', '_' * ((keylen - 5) + len(HEADER[-1])))
	# the suffix '40' is for the length restricted results
	maxlenseen = sentcount = maxlenseen40 = sentcount40 = 0
	goldb = multiset()
	candb = multiset()
	goldb40 = multiset()
	candb40 = multiset()
	goldbcat = defaultdict(multiset)
	candbcat = defaultdict(multiset)
	goldbcat40 = defaultdict(multiset)
	candbcat40 = defaultdict(multiset)
	lascores = []
	lascores40 = []
	golddep = []
	canddep = []
	golddep40 = []
	canddep40 = []
	goldpos = []
	candpos = []
	goldpos40 = []
	candpos40 = []
	exact = exact40 = 0.0
	dicenoms = dicedenoms = dicenoms40 = dicedenoms40 = 0
	for n, ctree in cand_trees.items():
		gtree = gold_trees[n]
		cpos = sorted(ctree.pos())
		gpos = sorted(gtree.pos())
		csent = [w for w, _ in cand_sents[n]]
		gsent = [w for w, _ in gold_sents[n]]
		lencpos = sum(1 for _, b in cpos
			if b not in param['DELETE_LABEL_FOR_LENGTH'])
		lengpos = sum(1 for _, b in gpos
			if b not in param['DELETE_LABEL_FOR_LENGTH'])
		assert lencpos == lengpos, ('sentence length mismatch. '
				'sents:\n%s\n%s' % (' '.join(csent), ' '.join(gsent)))
		# massage the data (in-place modifications)
		transform(ctree, csent, cpos, dict(gpos), param['DELETE_LABEL'],
				param['DELETE_WORD'], param['EQ_LABEL'], param['EQ_WORD'])
		transform(gtree, gsent, gpos, dict(gpos), param['DELETE_LABEL'],
				param['DELETE_WORD'], param['EQ_LABEL'], param['EQ_WORD'])
		#if not gtree or not ctree:
		#	continue
		assert csent == gsent, ('candidate & gold sentences do not match:\n'
			'%r // %r' % (' '.join(csent), ' '.join(gsent)))
		cbrack = bracketings(ctree, param['LABELED'], param['DELETE_LABEL'],
				param['DISC_ONLY'])
		gbrack = bracketings(gtree, param['LABELED'], param['DELETE_LABEL'],
				param['DISC_ONLY'])
		if not param['DISC_ONLY'] or cbrack or gbrack:
			sentcount += 1
		# this is to deal with 'sentences' with only a single punctuation mark.
		if not gpos:
			continue
		if maxlenseen < lencpos:
			maxlenseen = lencpos
		if cbrack == gbrack:
			if not param['DISC_ONLY'] or cbrack or gbrack:
				exact += 1
		candb.update((n, a) for a in cbrack.elements())
		goldb.update((n, a) for a in gbrack.elements())
		for a in gbrack:
			goldbcat[a[0]][(n, a)] += 1
		for a in cbrack:
			candbcat[a[0]][(n, a)] += 1
		goldpos.extend(gpos)
		candpos.extend(cpos)
		lascores.append(leafancestor(gtree, ctree, param['DELETE_LABEL']))
		if param['TED']:
			ted, denom = treedisteval(gtree, ctree,
				includeroot=gtree.label not in param['DELETE_LABEL'])
			dicenoms += ted
			dicedenoms += denom
		if param['DEP']:
			cdep = dependencies(ctree, param['HEADRULES'])
			gdep = dependencies(gtree, param['HEADRULES'])
			canddep.extend(cdep)
			golddep.extend(gdep)
		if lencpos <= param['CUTOFF_LEN']:
			if not param['DISC_ONLY'] or cbrack or gbrack:
				sentcount40 += 1
			if maxlenseen40 < lencpos:
				maxlenseen40 = lencpos
			candb40.update((n, a) for a in cbrack.elements())
			goldb40.update((n, a) for a in gbrack.elements())
			for a in gbrack:
				goldbcat40[a[0]][(n, a)] += 1
			for a in cbrack:
				candbcat40[a[0]][(n, a)] += 1
			if cbrack == gbrack:
				if not param['DISC_ONLY'] or cbrack or gbrack:
					exact40 += 1
			goldpos40.extend(gpos)
			candpos40.extend(cpos)
			if lascores[-1] is not None:
				lascores40.append(lascores[-1])
			if param['TED']:
				dicenoms40 += ted
				dicedenoms40 += denom
			if param['DEP']:
				canddep40.extend(cdep)
				golddep40.extend(gdep)
		assert lascores[-1] != 1 or gbrack == cbrack, (
				'leaf ancestor score 1.0 but no exact match: (bug?)')
		if lascores[-1] is None:
			del lascores[-1]
		if param['DEBUG'] <= 0:
			continue
		if param['DEBUG'] > 1:
			for a in HEADER:
				print(' ' * (keylen - 4) + a)
			print('', '_' * ((keylen - 5) + len(HEADER[-1])))
		print(('%' + str(keylen) +
			's  %5d  %s  %s   %5d  %5d  %5d  %5d  %4d  %s %6.2f%s%s') % (
				n,
				lengpos,
				nozerodiv(lambda: recall(gbrack, cbrack)),
				nozerodiv(lambda: precision(gbrack, cbrack)),
				sum((gbrack & cbrack).values()),
				sum(gbrack.values()),
				sum(cbrack.values()),
				len(gpos),
				sum(1 for a, b in zip(gpos, cpos) if a == b),
				nozerodiv(lambda: accuracy(gpos, cpos)),
				100 * lascores[-1],
				str(ted).rjust(3) if param['TED'] else '',
				nozerodiv(lambda: accuracy(gdep, cdep))
						if param['DEP'] else ''))
		if param['DEBUG'] > 1:
			print('Sentence:', ' '.join(gsent))
			print('Gold tree:\n%s\nCandidate tree:\n%s' % (
					DrawTree(gtree, gsent, abbr=True).text(
						unicodelines=True, ansi=True),
					DrawTree(ctree, csent, abbr=True).text(
						unicodelines=True, ansi=True)))
			print('Gold brackets:      %s\nCandidate brackets: %s' % (
					strbracketings(gbrack), strbracketings(cbrack)))
			print('Matched brackets:      %s\nUnmatched brackets: %s' % (
					strbracketings(gbrack & cbrack),
					strbracketings((cbrack - gbrack) | (gbrack - cbrack))))
			goldpaths = leafancestorpaths(gtree, param['DELETE_LABEL'])
			candpaths = leafancestorpaths(ctree, param['DELETE_LABEL'])
			for leaf in goldpaths:
				print('%6.3g  %s     %s : %s' % (
						pathscore(goldpaths[leaf], candpaths[leaf]),
						gsent[leaf].ljust(15),
						' '.join(goldpaths[leaf][::-1]).rjust(20),
						' '.join(candpaths[leaf][::-1])))
			print('%6.3g  average = leaf-ancestor score' % lascores[-1])
			print('POS: ', ' '.join('%s/%s' % (a[1], b[1])
					for a, b in zip(cpos, gpos)))
			if param['TED']:
				print('Tree-dist: %g / %g = %g' % (
					ted, denom, 1 - ted / denom))
				newtreedist(gtree, ctree, True)
			if param['DEP']:
				print('Sentence:', ' '.join(gsent))
				print('dependencies gold', ' ' * 35, 'cand')
				for (_, a, b), (_, c, d) in zip(gdep, cdep):
					# use original sentences because we don't delete
					# punctuation for dependency evaluation
					print('%15s -> %15s           %15s -> %15s' % (
						gold_sents[n][a - 1][0], gold_sents[n][b - 1][0],
						cand_sents[n][c - 1][0], cand_sents[n][d - 1][0]))
			print()
	breakdowns(param, goldb40, candb40, goldpos40, candpos40, goldbcat40,
			candbcat40, maxlenseen)
	msg = summary(param, goldb, candb, goldpos, candpos, sentcount,
			maxlenseen, exact, lascores, dicenoms, dicedenoms, golddep,
			canddep, goldb40, candb40, goldpos40, candpos40, sentcount40,
			maxlenseen40, exact40, lascores40, dicenoms40, dicedenoms40,
			golddep40, canddep40)
	return msg


def breakdowns(param, goldb, candb, goldpos, candpos, goldbcat, candbcat,
		maxlenseen):
	""" Print breakdowns for the most frequent labels / tags. """
	if param['LABELED'] and param['DEBUG'] != -1:
		print()
		print(' Category Statistics (10 most frequent categories / errors)',
				end='')
		if maxlenseen > param['CUTOFF_LEN']:
			print(' for length <= %d' % param['CUTOFF_LEN'], end='')
		print()
		print('  label  % gold  recall    prec.     F1',
				'          test/gold   count')
		print('_______________________________________',
				'       ____________________')
		gmismatch = {(n, indices): label
					for n, (label, indices) in (goldb - candb)}
		wrong = multiset((label, gmismatch[n, indices])
					for n, (label, indices) in (candb - goldb)
					if (n, indices) in gmismatch)
		freqcats = sorted(set(goldbcat) | set(candbcat),
				key=lambda x: len(goldbcat[x]), reverse=True)
		for cat, mismatch in zip_longest(freqcats[:10], wrong.most_common(10)):
			if cat is None:
				print('                                       ', end='')
			else:
				print('%s  %6.2f  %s  %s  %s' % (
					cat.rjust(7),
					100 * sum(goldbcat[cat].values()) / len(goldb),
					nozerodiv(lambda: recall(goldbcat[cat], candbcat[cat])),
					nozerodiv(lambda: precision(goldbcat[cat], candbcat[cat])),
					nozerodiv(lambda: f_measure(goldbcat[cat], candbcat[cat])),
					), end='')
			if mismatch is not None:
				print('       %s %7d' % (
						'/'.join(mismatch[0]).rjust(12), mismatch[1]), end='')
			print()

		if accuracy(goldpos, candpos) != 1:
			print()
			print(' Tag Statistics (10 most frequent tags / errors)', end='')
			if maxlenseen > param['CUTOFF_LEN']:
				print(' for length <= %d' % param['CUTOFF_LEN'], end='')
			print('\n    tag  % gold  recall   prec.      F1',
					'          test/gold   count')
			print('_______________________________________',
					'       ____________________')
			tags = multiset(tag for _, tag in goldpos)
			wrong = multiset((c, g)
					for (_, g), (_, c) in zip(goldpos, candpos) if g != c)
			for tag, mismatch in zip_longest(tags.most_common(10),
					wrong.most_common(10)):
				if tag is None:
					print(''.rjust(40), end='')
				else:
					goldtag = multiset(n for n, (w, t) in enumerate(goldpos)
							if t == tag[0])
					candtag = multiset(n for n, (w, t) in enumerate(candpos)
							if t == tag[0])
					print('%s  %6.2f  %6.2f  %6.2f  %6.2f' % (
							tag[0].rjust(7),
							100 * len(goldtag) / len(goldpos),
							100 * recall(goldtag, candtag),
							100 * precision(goldtag, candtag),
							100 * f_measure(goldtag, candtag)), end='')
				if mismatch is not None:
					print('       %s %7d' % (
						'/'.join(mismatch[0]).rjust(12), mismatch[1]), end='')
				print()
		print()


def summary(param, goldb, candb, goldpos, candpos, sentcount, maxlenseen,
		exact, lascores, dicenoms, dicedenoms, golddep, canddep,
		goldb40, candb40, goldpos40, candpos40, sentcount40, maxlenseen40,
		exact40, lascores40, dicenoms40, dicedenoms40, golddep40, canddep40):
	""" :returns: overview with scores for all sentences. """
	discbrackets = sum(1 for _, (_, a) in candb.elements()
			if a != tuple(range(min(a), max(a) + 1)))
	gdiscbrackets = sum(1 for _, (_, a) in goldb.elements()
			if a != tuple(range(min(a), max(a) + 1)))

	if maxlenseen <= param['CUTOFF_LEN']:
		msg = ['%s' % ' Summary (ALL) '.center(35, '_'),
			'number of sentences:       %6d' % (sentcount),
			'longest sentence:          %6d' % (maxlenseen)]
		if gdiscbrackets or discbrackets:
			msg.extend(['gold brackets (disc.):     %6d (%d)' % (
						len(goldb), gdiscbrackets),
					'cand. brackets (disc.):    %6d (%d)' % (
						len(candb), discbrackets)])
		else:
			msg.extend(['gold brackets:             %6d' % len(goldb),
				'cand. brackets:            %6d' % len(candb)])
		msg.extend([
				'labeled recall:            %s' % (
					nozerodiv(lambda: recall(goldb, candb))),
				'labeled precision:         %s' % (
					nozerodiv(lambda: precision(goldb, candb))),
				'labeled f-measure:         %s' % (
					nozerodiv(lambda: f_measure(goldb, candb))),
				'exact match:               %s' % (
					nozerodiv(lambda: exact / sentcount)),
				'leaf-ancestor:             %s' % (
					nozerodiv(lambda: mean(lascores)))])
		if param['TED']:
			msg.append('tree-dist (Dice micro avg) %s' % (
					nozerodiv(lambda: 1 - dicenoms / dicedenoms)))
		if param['DEP']:
			msg.append('unlabeled dependencies:    %s' % (
					nozerodiv(lambda: accuracy(golddep, canddep))))
		msg.append('tagging accuracy:          %s' % (
				nozerodiv(lambda: accuracy(goldpos, candpos))))
		return '\n'.join(msg)

	discbrackets40 = sum(1 for _, (_, a) in candb40.elements()
			if a != tuple(range(min(a), max(a) + 1)))
	gdiscbrackets40 = sum(1 for _, (_, a) in goldb40.elements()
			if a != tuple(range(min(a), max(a) + 1)))
	msg = ['%s <= %d ______ ALL' % (
			' Summary '.center(27, '_'), param['CUTOFF_LEN']),
		'number of sentences:       %6d     %6d' % (sentcount40, sentcount),
		'longest sentence:          %6d     %6d' % (maxlenseen40, maxlenseen),
		'gold brackets:             %6d     %6d' % (len(goldb40), len(goldb)),
		'cand. brackets:            %6d     %6d' % (len(candb40), len(candb))]
	if gdiscbrackets or discbrackets:
		msg.extend(['disc. gold brackets:       %6d     %6d' % (
				gdiscbrackets40, gdiscbrackets),
				'disc. cand. brackets:      %6d     %6d' % (
				discbrackets40, discbrackets)])
	msg.extend(['labeled recall:            %s     %s' % (
			nozerodiv(lambda: recall(goldb40, candb40)),
			nozerodiv(lambda: recall(goldb, candb))),
		'labeled precision:         %s     %s' % (
			nozerodiv(lambda: precision(goldb40, candb40)),
			nozerodiv(lambda: precision(goldb, candb))),
		'labeled f-measure:         %s     %s' % (
			nozerodiv(lambda: f_measure(goldb40, candb40)),
			nozerodiv(lambda: f_measure(goldb, candb))),
		'exact match:               %s     %s' % (
			nozerodiv(lambda: exact40 / sentcount40),
			nozerodiv(lambda: exact / sentcount)),
		'leaf-ancestor:             %s     %s' % (
			nozerodiv(lambda: mean(lascores40)),
			nozerodiv(lambda: mean(lascores)))])
	if param['TED']:
		msg.append('tree-dist (Dice micro avg) %s     %s' % (
			nozerodiv(lambda: (1 - dicenoms40 / dicedenoms40)),
			nozerodiv(lambda: (1 - dicenoms / dicedenoms))))
	if param['DEP']:
		msg.append('unlabeled dependencies:    %s     %s  (%d / %d)' % (
				nozerodiv(lambda: accuracy(golddep40, canddep40)),
				nozerodiv(lambda: accuracy(golddep, canddep)),
				len([a for a in zip(golddep, canddep) if a[0] == a[1]]),
				len(golddep)))
	msg.append('tagging accuracy:          %s     %s' % (
			nozerodiv(lambda: accuracy(goldpos40, candpos40)),
			nozerodiv(lambda: accuracy(goldpos, candpos))))
	return '\n'.join(msg)


def readparam(filename):
	""" read an EVALB-style parameter file and return a dictionary. """
	param = defaultdict(list)
	# NB: we ignore MAX_ERROR, we abort immediately on error.
	validkeysonce = ('DEBUG', 'MAX_ERROR', 'CUTOFF_LEN', 'LABELED',
			'DISC_ONLY', 'TED', 'DEP')
	param = {'DEBUG': 0, 'MAX_ERROR': 10, 'CUTOFF_LEN': 40,
				'LABELED': 1, 'DELETE_LABEL_FOR_LENGTH': set(),
				'DELETE_LABEL': set(), 'DELETE_WORD': set(),
				'EQ_LABEL': set(), 'EQ_WORD': set(),
				'DISC_ONLY': 0, 'TED': 0, 'DEP': 0}
	seen = set()
	for a in io.open(filename, encoding='utf8') if filename else ():
		line = a.strip()
		if line and not line.startswith('#'):
			key, val = line.split(None, 1)
			if key in validkeysonce:
				assert key not in seen, 'cannot declare %s twice' % key
				seen.add(key)
				param[key] = int(val)
			elif key in ('DELETE_LABEL', 'DELETE_LABEL_FOR_LENGTH',
					'DELETE_WORD'):
				param[key].add(val)
			elif key in ('EQ_LABEL', 'EQ_WORD'):
				# these are given as undirected pairs (A, B), (B, C), ...
				# to be represented as equivalence classes A => {A, B, C, D}
				try:
					b, c = val.split()
				except ValueError:
					raise ValueError('%s requires two values' % key)
				param[key].add((b, c))
			else:
				raise ValueError('unrecognized parameter key: %s' % key)
	for key in ('EQ_LABEL', 'EQ_WORD'):
		# from arbitrary pairs: [('A', 'B'), ('B', 'C')]
		# to eq classes: {'A': {'A', 'B', 'C'}}
		# to a mapping of all elements to their representative:
		# {'A': 'A', 'B': 'A', 'C': 'A'}
		param[key] = {x: k
				for k, eqclass in transitiveclosure(param[key]).items()
					for x in eqclass}
	return param


def transitiveclosure(eqpairs):
	""" Transitive closure of (undirected) EQ relations with DFS;
	i.e., given a sequence of pairs denoting an equivalence relation,
	produce a dictionary with equivalence classes as values and
	arbitrary members of those classes as keys.
	>>> result = transitiveclosure({('A', 'B'), ('B', 'C')})
	>>> len(result)
	1
	>>> k, v = result.popitem()
	>>> k in ('A', 'B', 'C') and v == {'A', 'B', 'C'}
	True """
	edges = defaultdict(set)
	for a, b in eqpairs:
		edges[a].add(b)
		edges[b].add(a)
	eqclasses = {}
	seen = set()
	for elem in set(edges):
		if elem in seen:
			continue
		eqclasses[elem] = set()
		agenda = edges.pop(elem)
		while agenda:
			eqelem = agenda.pop()
			seen.add(eqelem)
			eqclasses[elem].add(eqelem)
			agenda.update(edges[eqelem] - seen)
	return eqclasses


def transform(tree, sent, pos, gpos, dellabel, delword, eqlabel, eqword):
	""" Apply the transformations according to the parameter file,
	except for deleting the root node, which is a special case because if there
	is more than one child it cannot be deleted.

	:param pos: a list with the contents of tree.pos(); modified in-place.
	:param gpos: a dictionary of the POS tags of the original gold tree, before
		any tags/words have been deleted. """
	leaves = list(range(len(sent)))
	posnodes = []
	for a in reversed(list(tree.subtrees(lambda n: isinstance(n[0], Tree)))):
		for n, b in list(zip(count(), a))[::-1]:
			b.label = eqlabel.get(b.label, b.label)
			if not b:
				a.pop(n)  # remove empty nodes
			elif isinstance(b[0], Tree):
				if b.label in dellabel:
					# replace phrasal node with its children
					# (must remove nodes from b first because ParentedTree)
					bnodes = b[:]
					b[:] = []
					a[n:n + 1] = bnodes
			elif gpos[b[0]] in dellabel or sent[b[0]] in delword:
				# remove pre-terminal entirely, but only look at gold tree,
				# to ensure the sentence lengths stay the same
				leaves.remove(b[0])
				del a[n]
			else:
				posnodes.append(b)
				a.indices = b.indices = b[:]
	# retain words still in tree
	sent[:] = [sent[n] for n in leaves]
	pos[:] = [pos[n] for n in leaves]
	# removed POS tags cause the numbering to be off, restore.
	leafmap = {m: n for n, m in enumerate(leaves)}
	for a in posnodes:
		a[0] = leafmap[a[0]]
		a.indices = [a[0]]
		if sent[a[0]] in eqword:
			sent[a[0]] = eqword[sent[a[0]]]
	# cache spans
	for a in reversed(list(tree.subtrees())):
		indices = []
		for b in a:
			if isinstance(b, Tree):
				indices.extend(b.indices)
			elif isinstance(b, int):
				indices.append(b)
			else:
				raise ValueError('Tree should consist of Tree nodes and '
						'integer indices:\n%r' % b)
		assert len(indices) == len(set(indices)), (
			'duplicate index in tree:\n%s' % tree)
		a.indices = tuple(sorted(indices))


def bracketings(tree, labeled=True, dellabel=(), disconly=False):
	""" :returns: the labeled set of bracketings for a tree

	For each nonterminal node, the set will contain a tuple with the label and
	the set of terminals which it dominates.
	``tree`` must have been processed by ``transform()``.
	The argument ``dellabel`` is only used to exclude the ROOT node from the
	results (because it cannot be deleted by ``transform()`` when non-unary).

	>>> tree = Tree.parse('(S (NP 1) (VP (VB 0) (JJ 2)))', parse_leaf=int)
	>>> transform(tree, tree.leaves(), tree.pos(), dict(tree.pos()), (),
	... (), {}, {})
	>>> sorted(bracketings(tree).items())
	[(('S', (0, 1, 2)), 1), (('VP', (0, 2)), 1)]
	>>> tree = Tree.parse('(S (NP 1) (VP (VB 0) (JJ 2)))', parse_leaf=int)
	>>> transform(tree, tree.leaves(), tree.pos(), dict(tree.pos()), ('VP',),
	... (), {}, {})
	>>> bracketings(tree)
	Counter({('S', (0, 1, 2)): 1})
	>>> tree = Tree.parse('(S (NP 1) (VP (VB 0) (JJ 2)))', parse_leaf=int)
	>>> transform(tree, tree.leaves(), tree.pos(), dict(tree.pos()), ('S',),
	... (), {}, {})
	>>> bracketings(tree, dellabel=('S',))
	Counter({('VP', (0, 2)): 1})
	"""
	return multiset(bracketing(a, labeled) for a in tree.subtrees()
			if a and isinstance(a[0], Tree)  # nonempty, not a preterminal
				and a.label not in dellabel
				and (not disconly or disc(a)))


def bracketing(node, labeled=True):
	""" Generate bracketing ``(label, indices)`` for a given node. """
	return (node.label if labeled else '', node.indices)


def strbracketings(brackets):
	""" :returns: a string with a concise representation of a bracketing.

	>>> strbracketings({('S', (0, 1, 2)), ('VP', (0, 2))})
	'S[0-2], VP[0,2]'
	"""
	if not brackets:
		return '{}'
	return ', '.join('%s[%s]' % (a, ','.join(
		'-'.join(str(y) for y in sorted(set(x)))
		for x in intervals(sorted(b)))) for a, b in sorted(brackets))


def leafancestorpaths(tree, dellabel):
	""" Generate a list of ancestors for each leaf node in a tree. """
	#uses [] to mark components, and () to mark constituent boundaries
	#deleted words/tags should not affect boundary detection
	paths = {a: [] for a in tree.indices}
	# do a top-down level-order traversal
	thislevel = [tree]
	while thislevel:
		nextlevel = []
		for n in thislevel:
			leaves = sorted(n.indices)
			# skip empty nodes and POS tags
			if not leaves or not isinstance(n[0], Tree):
				continue
			first, last = min(leaves), max(leaves)
			# skip root node if it is to be deleted
			if n.label not in dellabel:
				for b in leaves:
					# mark end of constituents / components
					if b + 1 not in leaves:
						if b == last and ')' not in paths[b]:
							paths[b].append(')')
						elif b != last and ']' not in paths[b]:
							paths[b].append(']')
					# add this label to the lineage
					paths[b].append(n.label)
					# mark beginning of constituents / components
					if b - 1 not in leaves:
						if b == first and '(' not in paths[b]:
							paths[b].append('(')
						elif b != first and '[' not in paths[b]:
							paths[b].append('[')
			nextlevel.extend(n)
		thislevel = nextlevel
	return paths


def pathscore(gold, cand):
	""" Get edit distance for two leaf-ancestor paths. """
	return 1.0 - (edit_distance(cand, gold)
					/ max(len(gold) + len(cand), 1))


def leafancestor(goldtree, candtree, dellabel):
	""" Geoffrey Sampson, Anna Babarcz (2003):
	A test of the leaf-ancestor metric for parse accuracy """
	gold = leafancestorpaths(goldtree, dellabel)
	cand = leafancestorpaths(candtree, dellabel)
	return mean([pathscore(gold[leaf], cand[leaf]) for leaf in gold])


def treedisteval(a, b, includeroot=False, debug=False):
	""" Get tree-distance for two trees and compute the Dice normalization. """
	ted = treedist(a, b, debug)
	# Dice denominator
	denom = len(list(a.subtrees()) + list(b.subtrees()))
	# optionally discount ROOT nodes and preterminals
	if not includeroot:
		denom -= 2
	#if not includepreterms:
	#	denom -= len(a.leaves() + b.leaves())
	return ted, denom


# If the goldfile contains n constituents for the same span, and the parsed
# file contains m constituents with that nonterminal, the scorer works as
# follows:
#
# i) If m>n, then the precision is n/m, recall is 100%
#
# ii) If n>m, then the precision is 100%, recall is m/n.
#
# iii) If n==m, recall and precision are both 100%.
def recall(reference, candidate):
	""" Get recall score for two multisets. """
	if not reference:
		return float('nan')
	return sum(min(reference[a], candidate[a])
			for a in reference & candidate) / sum(reference.values())


def precision(reference, candidate):
	""" Get precision score for two multisets. """
	if not candidate:
		return float('nan')
	return sum(min(reference[a], candidate[a])
			for a in reference & candidate) / sum(candidate.values())


def f_measure(reference, candidate, alpha=0.5):
	""" Get F-measure of precision and recall for two multisets.
	The default weight alpha=0.5 corresponds to the F_1-measure. """
	p = precision(reference, candidate)
	r = recall(reference, candidate)
	if p == 0 or r == 0:
		return float('nan')
	return 1.0 / (alpha / p + (1 - alpha) / r)


def accuracy(reference, candidate):
	""" Given a sequence of reference values and a corresponding sequence of
	test values, return the fraction of corresponding values that are equal.
	In particular, return the fraction of indices
	0<i<=len(test) such that test[i] == reference[i]. """
	assert len(reference) == len(candidate), (
		'Sequences must have the same length.')
	return sum(1 for a, b in zip(reference, candidate)
			if a == b) / len(reference)


def harmean(seq):
	""" Compute harmonic mean of a sequence of non-zero numbers. """
	numerator = denominator = 0
	for a in seq:
		if not a:
			return float('nan')
		numerator += 1
		denominator += 1 / a
	if not denominator:
		return float('nan')
	return numerator / denominator


def mean(seq):
	""" Compute arithmetic mean of a sequence. """
	numerator = denominator = 0
	for a in seq:
		numerator += a
		denominator += 1
	if not denominator:
		return float('nan')
	return numerator / denominator


def intervals(seq):
	""" Partition seq into a sequence of intervals corresponding to contiguous
	ranges. An interval is a pair (a, b), with a <= b denoting terminals x
	such that a <= x <= b.

	>>> list(intervals((0, 1, 3, 4, 6, 7, 8)))
	[(0, 1), (3, 4), (6, 8)] """
	start = prev = None
	for a in seq:
		if start is None:
			start = prev = a
		elif a == prev + 1:
			prev = a
		else:
			yield start, prev
			start = prev = a
	if start is not None:
		yield start, prev


def disc(node):
	""" This function evaluates whether a particular node is locally
	discontinuous. The root node will, by definition, be continuous.
	Nodes can be continuous even if some of their children are discontinuous.
	"""
	if not isinstance(node, Tree):
		return False
	start = prev = None
	for a in sorted(node.indices):
		if start is None:
			start = prev = a
		elif a == prev + 1:
			prev = a
		else:
			return True
	return False


def nozerodiv(a):
	""" Convenience function to catch zero division or None as a result,
	and otherwise format as a percentage with two decimals. """
	try:
		result = a()
	except ZeroDivisionError:
		return ' 0DIV!'
	return '  None' if result is None else '%6.2f' % (100 * result)


def edit_distance(seq1, seq2):
	""" Calculate the Levenshtein edit-distance between two strings. The edit
	distance is the number of characters that need to be substituted, inserted,
	or deleted, to transform seq1 into seq2.  For example, transforming 'rain'
	to 'shine' requires three steps, consisting of two substitutions and one
	insertion: 'rain' -> 'sain' -> 'shin' -> 'shine'.  These operations could
	have been done in other orders, but at least three steps are needed.
	"""
	# set up a 2-D array
	len1 = len(seq1)
	len2 = len(seq2)
	# initialize 2-D array to zero
	lev = [[0] * (len2 + 1) for _ in range(len1 + 1)]
	for i in range(len1 + 1):
		lev[i][0] = i           # column 0: 0,1,2,3,4,...
	for j in range(len2 + 1):
		lev[0][j] = j           # row 0: 0,1,2,3,4,...

	# iterate over the array
	for i in range(len1):
		for j in range(len2):
			a = lev[i][j + 1] + 1               # skip seq1[i]
			b = lev[i][j] + (seq1[i] != seq2[j])  # match seq1[i] with seq2[j]
			c = lev[i + 1][j] + 1               # skip seq2[j]
			lev[i + 1][j + 1] = min(a, b, c)    # pick the cheapest
	return lev[len1][len2]


def test():
	""" Simple sanity check; should give 100% score on all metrics. """
	gold = getreader('export')('.', 'alpinosample.export')
	parses = getreader('export')('.', 'alpinosample.export')
	doeval(gold.parsed_sents(),
			gold.tagged_sents(),
			parses.parsed_sents(),
			parses.tagged_sents(),
			readparam(None))

if __name__ == '__main__':
	main()



from math import isinf, exp, log, fsum
from libc.stdlib cimport malloc, realloc, free
from libc.string cimport memcmp
cimport cython

ctypedef unsigned long long ULLong
ctypedef unsigned long ULong
ctypedef unsigned int UInt
ctypedef unsigned char UChar


cdef extern:
	int __builtin_ffsll (ULLong)
	int __builtin_ctzll (ULLong)
	int __builtin_clzll (ULLong)
	int __builtin_ctzl (ULong)
	int __builtin_popcountl (ULong)
	int __builtin_popcountll (ULLong)


cdef extern from "macros.h":
	int BITSIZE
	int BITSLOT(int b)
	ULong BITMASK(int b)
	int BITNSLOTS(int nb)
	void SETBIT(ULong a[], int b)
	ULong TESTBIT(ULong a[], int b)
	#int SLOTS # doesn't work


DEF SLOTS = 3 # FIXME: make this a constant, yet shared across modules.


cdef struct Rule:
	double prob # 8 bytes
	UInt args # 4 bytes => 32 max vars per rule
	UInt lengths # 4 bytes => same
	UInt lhs # 4 bytes
	UInt rhs1 # 4 bytes
	UInt rhs2 # 4 bytes
	UInt no # 4 bytes
	# total: 32 bytes.


@cython.final
cdef class LexicalRule:
	cdef UInt lhs
	cdef unicode word
	cdef double prob


cdef class ChartItem:
	cdef UInt label


@cython.final
cdef class SmallChartItem(ChartItem):
	cdef ULLong vec


@cython.final
cdef class FatChartItem(ChartItem):
	cdef ULong vec[SLOTS]


@cython.final
cdef class CFGChartItem(ChartItem):
	cdef UChar start, end


cdef SmallChartItem CFGtoSmallChartItem(UInt label, UChar start, UChar end)
cdef FatChartItem CFGtoFatChartItem(UInt label, UChar start, UChar end)


cdef class Edge:
	cdef double inside
	cdef Rule *rule


@cython.final
cdef class LCFRSEdge(Edge):
	cdef double score # inside probability + estimate score
	cdef ChartItem left
	cdef ChartItem right


@cython.final
cdef class CFGEdge(Edge):
	cdef UChar mid


@cython.final
cdef class RankedEdge:
	cdef ChartItem head
	cdef LCFRSEdge edge
	cdef int left
	cdef int right


@cython.final
cdef class RankedCFGEdge:
	cdef UInt label
	cdef UChar start, end
	cdef CFGEdge edge
	cdef int left
	cdef int right


# start scratch
#cdef class ParseForest:
#	""" the chart representation of bitpar. seems to require parsing
#	in 3 stages: recognizer, enumerate analyses, get probs. """
#	#keys
#	cdef UInt *catnum			# no. of chart item -> lhs
#	cdef size_t *firstanalysis	# no. of chart item -> idx to arrays below.
#	# from firstanalysis[n] to firstanalysis[n+1] or end values.
#	cdef size_t *firstchild     # idx to child array below
#	cdef double *insideprobs	# no. of edge -> inside prob
#	cdef UInt *ruleno
#	#positive means index to lists above, negative means terminal index
#	cdef UInt *child
#
# instead of explicitly recording ruleno & mid, record them as part of
# indices / hashes
#cdef struct PackedCFGEdge: # 17 bytes
#	# float here could help squeeze this in 13 bytes
#	double prob
#	# on 64 bits, UInt for rule number is as big as a pointer,
#	# so might as well store pointer directly.
#	Rule *rule
#	UChar mid
#
#cdef class ParseForest:
#	# each labeled span is a separately allocated array of edges
#	PackedCFGEdge **edges
#	size_t *celloffsets, numedges, arraysize
#	UInt nonterminals
#	UChar lensent
#	def __init__(self, lensent, nonterminals):
#		self.lensent = lensent
#		self.nonterminals = nonterminals
#		self.numedges = 0
#		cells = lensent * (lensent + 1) / 2
#		self.celloffsets = <size_t *>malloc(cells * sizeof(size_t))
#		# NB: first instance of 'nonterminals' is for POS tags
#		self.arraysize = (lensent * nonterminals
#				+ (cells - lensent) * nonterminals)
#		self.edges = <PackedCFGEdge **>malloc(
#				self.arraysize * sizeof(*PackedCFGEdge))
#		for start in range(lensent):
#			for end in range(start + 1, lensent + 1):
#				self.celloffsets[self.cellindex(start, end)] = (
#						self.nonterminals * self.cellindex(start, end))
#	cdef cellindex(UChar start, UChar end):
#		return (self.lensent * start - ((start - 1) * start / 2)
#				+ end - start - 1)
#	cdef PackedCFGEdge get(start, end, label, n):
#		return self.edges[(left * self.lensent + right) * lensent + label][n]
#	cdef set(start, end, label, ruleno, mid, prob)
#		cdef size_t idx = (left * self.lensent + right) * lensent + label
#		n = self.numedges[idx]
#		self.edges[idx][n].ruleno = ruleno
#		self.edges[idx][n].mid = mid
#		self.edges[idx][n].prob = prob
#		self.numedges += 1
#
#cdef class PackedParseForest:
#	# all edges listed contiguously in a single array
#	PackedCFGEdge *edges  # dynamically grown
#	size_t numedges, allocatededges
#	size_t *celloffset  # size depends only on sent length
#	size_t *cellpopulation  # number of edges in cell
#	#celloffset[left * lensent + right] => idx to edges of cell, sorted by lhs
#	# offset for lhs obtained with binary search
#	def __init__(self, lensent, nonterminals):
#		self.lensent = lensent
#		self.nonterminals = nonterminals
#		self.allocatededges = 10000
#		self.edges = alloc(self.edges, self.allocatededges * sizeof(CFGEdge))
#	cdef size_t getoffset(UInt label, UChar start, UChar end)
#		# do binary search for lhs at celloffset[start, end]
#		return binarysearch(label, edges, celloffset[start, end],
#				celloffset[start, end] + cellpopulation[start, end])
#	cdef inline addedge(self, size_t celloffset,
#			left, right, mid, lhs, rule, inside):
#		# NB: binary edges can be added in order,
#		# use sort after adding unaries.
#		cdef PackedCFGEdge *newedge
#		#if ...: # already in chart
#		#	# does this happen for binary rules? probably not.
#		#	# for unary rules? avoid.
#		if numedges == allocatededges:
#			self.realloc()
#		newedge = &(self.edges[celloffset
#				+ self.cellpopulation[celloffset]])
#		newedge.ruleno = ruleno
#		newedge.inside = inside
#		newedge.mid = mid
#		self.numedges += 1
#		self.cellpopulation[left, right] += 1
#	cdef void realloc(self):
#		self.allocatededges *= 2
#		self.edges = realloc(self.edges, self.allocatededges * sizeof(CFGEdge))
#		assert self.edges is not NULL
#
#cdef inline PackedCFGEdge getviterbi(PackedParseForest forest,
#		size_t celloffset, left, right, lhs):
#	return forest.edges[celloffset + forest.labeloffset[celloffset][lhs]]
#cdef inline PackedCFGEdge getedge(PackedParseForest forest, size_t celloffset,
#		left, right, mid, ruleno, inside):
#	return forest.edges[celloffset + forest.edgeoffset[celloffset][mid][ruleno]]
#
#cdef class ParseForest:
#	""" A packed parse forest represented in contiguous arrays.
#	This only works if for each cell, all rules for each lhs are added
#	in order. """
#	cdef PackedCFGEdge **edges
#	#
#	cdef UInt *rules
#	cdef short *midpoints
#	cdef double *inside
#	# offset[lhs] contains the index where the edges for that label start.
#	cdef size_t *offset
#	# unaries are stored separately:
#	# unaryrules[left * lensent + right][edgeno]
#	# unaryinside[left * lensent + right]edgeno]
#	cdef Rule **unaryrules
#	cdef double **unaryinside
#	#cdef __init__(self, Grammar grammar, short lensent):
#	#	self.lensent = lensent
#	#	self.numrules = grammar.numrules
#	#	# this is for a dense array,
#	#	# start with conservative estimate and then re-alloc along the way?
#	#	self.midpoints = calloc(sizeof(self.midpoints), lensent * lensent)
#	#	self.rules = calloc(sizeof(self.rules), lensent * lensent)
#	#	self.inside = calloc(sizeof(self.inside), lensent * lensent)
#	#	self.offset = calloc(sizeof(self.offset), len(grammar.tolabel))
#	#	self.unaryrules = calloc(sizeof(self.unaryrules), lensent * lensent
#	#			* len(grammar.numunary))
#	#	self.unaryinside = calloc(sizeof(self.unaryinside), lensent * lensent
#	#			* len(grammar.numunary))
#	cdef inline getinside(self, short left, short right, UInt label):
#		cdef size_t dest = self.offsets[label + 1]
#		return self.inside[left * self.lensent + right][self.offset[label]]
#	cdef inline getrule(self, short left, short right, UInt label):
#		return self.rules[left * self.lensent + right][self.offset[label]]
#	cdef inline getmidpoint(self, short left, short right, UInt label):
#		return self.midpoints[left * self.lensent + right][self.offset[label]]
#	cdef inline setitem(self, short left, short right, UInt label,
#			UInt ruleno, short mid, double inside):
#		cdef size_t dest = self.offsets[label + 1]
#		self.rules[dest] = ruleno
#		self.midpoints[dest] = mid
#		self.inside[dest] = inside
#		self.offsets[label + 1] += 1
#	#cdef __dealloc__(self):
#	#	if self.inside is not None:
#	#		free(self.inside)
#	#		free(self.rules)
#	#		free(self.midpoints)
#	#		self.inside = None
#	#		self.rules = None
#	#		self.midpoints = None
#
#cdef class DiscNode:
#	cdef int label
#	cdef tuple children
#	cdef CBitset leaves
# end scratch

# start fragments stuff

cdef struct Node:  # a node of a binary tree
	int prod # non-negative, ID of a phrasal or lexical production
	short left # >= 0: array idx to child Node; <0: idx sent[-left - 1];
	short right # >=0: array idx to child Node; -1: empty (unary Node)


cdef struct NodeArray:  # a tree as an array of Node structs
	size_t offset # index to array of nodes in treebank where this tree starts
	short len, root # number of nodes, index to root node


@cython.final
cdef class Ctrees:
	cdef Node *nodes
	cdef NodeArray *trees
	cdef long nodesleft, max
	cdef readonly size_t numnodes
	cdef readonly short maxnodes
	cdef readonly int len
	cdef list treeswithprod
	cpdef alloc(self, int numtrees, long numnodes)
	cdef realloc(self, int len)
	cpdef add(self, list tree, dict prods)
	cdef addnodes(self, Node *source, int cnt, int root)

# end fragments stuff


# ---------------------------------------------------------------
#                          INLINED FUNCTIONS
# ---------------------------------------------------------------


cdef inline FatChartItem new_FatChartItem(UInt label):
	cdef FatChartItem item = FatChartItem.__new__(FatChartItem)
	item.label = label
	# NB: since item.vec is a static array, its elements are initialized to 0.
	return item


cdef inline SmallChartItem new_ChartItem(UInt label, ULLong vec):
	cdef SmallChartItem item = SmallChartItem.__new__(SmallChartItem)
	item.label = label
	item.vec = vec
	return item


cdef inline CFGChartItem new_CFGChartItem(UInt label, UChar start, UChar end):
	cdef CFGChartItem item = CFGChartItem.__new__(CFGChartItem)
	item.label = label
	item.start = start
	item.end = end
	return item


cdef inline LCFRSEdge new_LCFRSEdge(double score, double inside, Rule *rule,
		ChartItem left, ChartItem right):
	cdef LCFRSEdge edge = LCFRSEdge.__new__(LCFRSEdge)
	cdef long h = 0x345678UL
	edge.score = score
	edge.inside = inside
	edge.rule = rule
	edge.left = left
	edge.right = right
	return edge


cdef inline CFGEdge new_CFGEdge(double inside, Rule *rule, UChar mid):
	cdef CFGEdge edge = CFGEdge.__new__(CFGEdge)
	edge.inside = inside
	edge.rule = rule
	edge.mid = mid
	return edge


cdef object log1e200 = log(1e200)


cdef inline logprobadd(x, y):
	""" Add two log probabilities in log space; i.e.:

	>>> a = b = 0.25
	>>> logprobadd(log(a), log(b)) == log(a + b) == log(0.5)
	True

	:param x, y: Python floats with log probabilities; -inf <= x, y <= 0.
	:source: https://facwiki.cs.byu.edu/nlp/index.php/Log_Domain_Computations
	"""
	if isinf(x):
		return y
	elif isinf(y):
		return x
	# If one value is much smaller than the other, keep the larger value.
	elif x < (y - 460):  # log(1e200)
		return y
	elif y < (x - 460):  # log(1e200)
		return x
	diff = y - x
	assert not isinf(diff)
	if isinf(exp(diff)):	# difference is too large
		return x if x > y else y
	# otherwise return the sum.
	return x + log(1.0 + exp(diff))


cdef inline double logprobsum(list logprobs):
	""" Takes a list of log probabilities and sums them producing a
	normal probability 0 < p <= 1.0; i.e.:

	>>> a = b = c = 0.25
	>>> logprobsum([log(a), log(b), log(c)]) == sum([a, b, c]) == 0.75
	True

	:param logprobs: a list of Python floats with negative log probilities,
		s.t. 0 <= p <= inf for each p in ``logprobs``.
	:source: http://blog.smola.org/post/987977550/log-probabilities-semirings-and-floating-point-numbers

	Comparison of different methods: https://gist.github.com/andreasvc/6204982
	"""
	maxprob = max(logprobs)
	return exp(maxprob) * fsum([exp(prob - maxprob) for prob in logprobs])



""" Probabilistic Context-Free Grammar (PCFG) parser using CKY. """
# python imports
from __future__ import print_function
from math import exp, log as pylog
from collections import defaultdict
import numpy as np
from discodop.tree import Tree
from discodop.agenda import EdgeAgenda
# cython imports
from libc.stdlib cimport malloc, calloc, free
from cpython cimport PyDict_Contains, PyDict_GetItem
from discodop.agenda cimport EdgeAgenda
from discodop._grammar cimport Grammar
from discodop.containers cimport CFGEdge, CFGChartItem, new_CFGChartItem, \
		new_CFGEdge, Rule, LexicalRule, UChar, UInt, ULong, ULLong

cdef extern from "math.h":
	bint isinf(double x)
	bint isfinite(double x)

DEF SX = 1
DEF SXlrgaps = 2
cdef CFGChartItem NONE = new_CFGChartItem(0, 0, 0)


def parse(sent, Grammar grammar, tags=None, start=None, chart=None):
	""" Parse a sentence with a PCFG. Automatically decides whether to use
	parse_dense or parse_sparse. """
	assert grammar.maxfanout == 1, 'Not a PCFG! fanout: %d' % grammar.maxfanout
	assert grammar.logprob, "Expecting grammar with log probabilities."
	if grammar.nonterminals < 20000 and chart is None:
		return parse_dense(sent, grammar, start=start, tags=tags)
	return parse_sparse(sent, grammar, start=start, tags=tags, chart=chart)


def parse_dense(sent, Grammar grammar, start=None, tags=None):
	""" A CKY parser modeled after Bodenstab's 'fast grammar loop'
	and the Stanford parser. Tries to apply each grammar rule for all
	spans. Tracks the viterbi scores in a separate array. For grammars with
	up to 10,000 nonterminals.
	Edges are kept in a dictionary for each labelled span:
	chart[left][right][label] = {edge1: edge1, edge2: edge2} """
	cdef:
		short left, right, mid, span, lensent = len(sent)
		short narrowl, narrowr, widel, wider, minmid, maxmid
		UInt n, lhs, rhs1, spans = 0, items = 0, edges = 0
		double oldscore, prob
		Rule *rule
		LexicalRule lexrule
		EdgeAgenda unaryagenda = EdgeAgenda()
		list chart = [[{} for _ in range(lensent + 1)] for _ in range(lensent)]
		dict cell
		short [:, :] minleft, maxleft, minright, maxright
		double [:, :, :] viterbi
	viterbi = chartmatrix(grammar.nonterminals, lensent)
	assert len(viterbi) >= grammar.nonterminals
	assert len(viterbi[0]) >= lensent
	assert len(viterbi[0][0]) >= lensent + 1
	minleft, maxleft, minright, maxright = minmaxmatrices(
			grammar.nonterminals, lensent)
	if start is None:
		start = grammar.toid[grammar.start]
	for left, _ in enumerate(sent):
		chart[left][left + 1] = dict.fromkeys(range(1, grammar.nonterminals))
	viterbiedges, msg = populatepos(grammar, chart, sent, tags,
			minleft, maxleft, minright, maxright)
	if not viterbiedges:
		return chart, NONE, msg
	for left, viterbicell in enumerate(viterbiedges):
		for lhs, edge in viterbicell.items():
			viterbi[lhs, left, left + 1] = (<CFGEdge>edge).inside

	for span in range(2, lensent + 1):
		# constituents from left to right
		for left in range(lensent - span + 1):
			right = left + span
			cell = chart[left][right]
			# binary rules
			for n in range(grammar.numrules):
				rule = &(grammar.bylhs[0][n])
				if rule.lhs == grammar.nonterminals:
					break
				elif not rule.rhs2:
					continue
				lhs = rule.lhs
				narrowr = minright[rule.rhs1, left]
				if narrowr >= right:
					continue
				narrowl = minleft[rule.rhs2, right]
				if narrowl < narrowr:
					continue
				widel = maxleft[rule.rhs2, right]
				minmid = narrowr if narrowr > widel else widel
				wider = maxright[rule.rhs1, left]
				maxmid = wider if wider < narrowl else narrowl
				oldscore = viterbi[lhs, left, right]
				if isinf(oldscore):
					cell[lhs] = {}
				for mid in range(minmid, maxmid + 1):
					if (isfinite(viterbi[rule.rhs1, left, mid])
						and isfinite(viterbi[rule.rhs2, mid, right])):
						prob = (rule.prob + viterbi[rule.rhs1, left, mid]
								+ viterbi[rule.rhs2, mid, right])
						edge = new_CFGEdge(prob, rule, mid)
						if prob < viterbi[lhs, left, right]:
							if isinf(viterbi[lhs, left, right]):
								cell[lhs] = {}
							viterbi[lhs, left, right] = prob
							cell[lhs][edge] = edge
						elif (edge not in cell[lhs] or
								prob < (<CFGEdge>cell[lhs][edge]).inside):
							cell[lhs][edge] = edge
				# update filter
				if isinf(oldscore):
					if not cell[lhs]:
						del cell[lhs]
						continue
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right

			# unary rules on this span
			unaryagenda.update([(rhs1,
				new_CFGEdge(viterbi[rhs1, left, right], NULL, 0))
				for rhs1 in range(grammar.nonterminals)
				if isfinite(viterbi[rhs1, left, right])
				and grammar.unary[rhs1].rhs1 == rhs1])
			while unaryagenda.length:
				rhs1 = unaryagenda.popentry().key
				for n in range(grammar.numrules):
					rule = &(grammar.unary[rhs1][n])
					if rule.rhs1 != rhs1:
						break
					prob = rule.prob + viterbi[rhs1, left, right]
					lhs = rule.lhs
					edge = new_CFGEdge(prob, rule, right)
					if isfinite(viterbi[lhs, left, right]):
						if prob < viterbi[lhs, left, right]:
							unaryagenda.setifbetter(lhs, <CFGEdge>edge)
							viterbi[lhs, left, right] = prob
							cell[lhs][edge] = edge
						elif (edge not in cell[lhs] or
								prob < (<CFGEdge>cell[lhs][edge]).inside):
							cell[lhs][edge] = edge
						continue
					unaryagenda.setifbetter(lhs, <CFGEdge>edge)
					viterbi[lhs, left, right] = prob
					cell[lhs] = {edge: edge}
					# update filter
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right
			if cell:
				spans += 1
				items += len(cell)
				edges += sum(map(len, cell.values()))
	msg = "chart spans %d, items %d, edges %d" % (spans, items, edges)
	if not chart[0][lensent].get(start):
		return chart, NONE, "no parse " + msg
	return chart, new_CFGChartItem(start, 0, lensent), msg


def parse_sparse(sent, Grammar grammar, start=None, tags=None,
		list chart=None, int beamwidth=0):
	""" A CKY parser modeled after Bodenstab's 'fast grammar loop,' filtered by
	the list of allowed items (if a pre-populated chart is given).
	This version keeps the Viterbi probabilities and the rest of chart in
	hash tables, useful for large grammars. The edge with the viterbi score
	for a labeled span is kept in viterbi[left][right][label]. """
	cdef:
		short left, right, mid, span, lensent = len(sent)
		short narrowl, narrowr, widel, wider, minmid, maxmid
		UInt n, lhs, rhs1, spans = 0, items = 0, edges = 0
		double oldscore, prob, infinity = float('infinity')
		Rule *rule
		LexicalRule lexrule
		EdgeAgenda unaryagenda = EdgeAgenda()
		list viterbi = [[{} for _ in range(lensent + 1)]
				for _ in range(lensent)]
		dict cell, viterbicell
		short [:, :] minleft, maxleft, minright, maxright
	minleft, maxleft, minright, maxright = minmaxmatrices(
			grammar.nonterminals, lensent)
	if start is None:
		start = grammar.toid[grammar.start]
	if chart is None:
		chart = [[None] * (lensent + 1) for _ in range(lensent)]
		for left in range(lensent):
			for right in range(left, lensent):
				chart[left][right + 1] = {
						n: None for n in range(1, grammar.nonterminals)}
	# assign POS tags
	viterbiedges, msg = populatepos(grammar, chart, sent, tags,
			minleft, maxleft, minright, maxright)
	if not viterbiedges:
		return chart, NONE, msg
	for left, viterbicell in enumerate(viterbiedges):
		viterbi[left][left + 1] = viterbicell

	for span in range(2, lensent + 1):
		# constituents from left to right
		for left in range(lensent - span + 1):
			right = left + span
			viterbicell = viterbi[left][right]
			cell = chart[left][right]
			# binary rules
			for lhs in cell:
				n = 0
				rule = &(grammar.bylhs[lhs][n])
				oldscore = ((<CFGEdge>viterbicell[lhs]).inside
						if lhs in viterbicell else infinity)
				if not cell[lhs]:
					cell[lhs] = {}
				while rule.lhs == lhs:
					narrowr = minright[rule.rhs1, left]
					narrowl = minleft[rule.rhs2, right]
					if rule.rhs2 == 0 or narrowr >= right or narrowl < narrowr:
						n +=1
						rule = &(grammar.bylhs[lhs][n])
						continue
					widel = maxleft[rule.rhs2, right]
					minmid = narrowr if narrowr > widel else widel
					wider = maxright[rule.rhs1, left]
					maxmid = wider if wider < narrowl else narrowl
					for mid in range(minmid, maxmid + 1):
						if (rule.rhs1 in viterbi[left][mid]
								and rule.rhs2 in viterbi[mid][right]):
							prob = (rule.prob + (<CFGEdge>viterbi[
								left][mid][rule.rhs1]).inside + (<CFGEdge>
								viterbi[mid][right][rule.rhs2]).inside)
							edge = new_CFGEdge(prob, rule, mid)
							if (lhs not in viterbicell or
									prob < (<CFGEdge>viterbicell[lhs]).inside):
								viterbicell[lhs] = edge
								cell[lhs][edge] = edge
							elif edge not in cell[lhs]:
								cell[lhs][edge] = edge
					n +=1
					rule = &(grammar.bylhs[lhs][n])

				# update filter
				if isinf(oldscore):
					if not cell[lhs]:
						continue
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right

			# unary rules
			unaryagenda.update(viterbicell.items())
			while unaryagenda.length:
				rhs1 = unaryagenda.popentry().key
				for n in range(grammar.numrules):
					rule = &(grammar.unary[rhs1][n])
					if rule.rhs1 != rhs1:
						break
					elif rule.lhs not in cell:
						continue
					lhs = rule.lhs
					prob = rule.prob + (<CFGEdge>
							viterbicell[rhs1]).inside
					edge = new_CFGEdge(prob, rule, right)
					if cell[lhs]:
						if prob < (<CFGEdge>viterbicell[lhs]).inside:
							unaryagenda.setifbetter(lhs, <CFGEdge>edge)
							viterbicell[lhs] = edge
							cell[lhs][edge] = edge
						elif (edge not in cell[lhs] or
								prob < (<CFGEdge>cell[lhs][edge]).inside):
							cell[lhs][edge] = edge
						continue
					unaryagenda.setifbetter(lhs, <CFGEdge>edge)
					viterbicell[lhs] = edge
					cell[lhs] = {edge: edge}
					# update filter
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right
			nonempty = list(filter(None, cell.values()))
			if nonempty:
				spans += 1
				items += len(nonempty)
				# clean up labelled spans that were whitelisted but not used:
				for label in list(cell):
					if not cell[label]:
						del cell[label]
				edges += sum(map(len, nonempty))
	msg = "chart spans %d, items %d, edges %d" % (spans, items, edges)
	if not chart[0][lensent].get(start):
		return chart, NONE, "no parse " + msg
	return chart, new_CFGChartItem(start, 0, lensent), msg


def parse_symbolic(sent, Grammar grammar, start=None, tags=None):
	""" Parse sentence, a list of tokens, and produce a chart, either
	exhaustive or up until the first complete parse. Non-probabilistic version.

	:returns: a CFG chart with each 'probability' 1.
	:param start: integer corresponding to the start symbol that analyses
		should have, e.g., grammar.toid['ROOT']
	:param tags: optionally, a list with the corresponding POS tags
		for the words in sent. """
	cdef:
		short left, right, mid, span, lensent = len(sent)
		short narrowl, narrowr, widel, wider, minmid, maxmid
		UInt n, lhs, rhs1, spans = 0, items = 0, edges = 0
		bint newspan
		Rule *rule
		LexicalRule lexrule
		set unaryagenda = set()
		list chart = [[{} for _ in range(lensent + 1)] for _ in range(lensent)]
		dict cell
		short [:, :] minleft, maxleft, minright, maxright
	assert grammar.maxfanout == 1, "Not a PCFG! fanout: %d" % grammar.maxfanout
	minleft, maxleft, minright, maxright = minmaxmatrices(
			grammar.nonterminals, lensent)
	# assign POS tags
	viterbiedges, msg = populatepos(grammar, chart, sent, tags,
			minleft, maxleft, minright, maxright)
	if start is None:
		start = grammar.toid[grammar.start]
	if not viterbiedges:
		return chart, NONE, msg

	for span in range(2, lensent + 1):
		# constituents from left to right
		for left in range(lensent - span + 1):
			right = left + span
			cell = chart[left][right]
			# binary rules
			for n in range(grammar.numrules):
				rule = &(grammar.bylhs[0][n])
				if rule.lhs == grammar.nonterminals:
					break
				elif not rule.rhs2:
					continue
				lhs = rule.lhs
				narrowr = minright[rule.rhs1, left]
				if narrowr >= right:
					continue
				narrowl = minleft[rule.rhs2, right]
				if narrowl < narrowr:
					continue
				widel = maxleft[rule.rhs2, right]
				minmid = narrowr if narrowr > widel else widel
				wider = maxright[rule.rhs1, left]
				maxmid = wider if wider < narrowl else narrowl
				newspan = lhs not in cell
				if newspan:
					cell[lhs] = {}
				for mid in range(minmid, maxmid + 1):
					if (rule.rhs1 in chart[left][mid]
							and rule.rhs2 in chart[mid][right]):
						edge = new_CFGEdge(0.0, rule, mid)
						cell[lhs][edge] = edge

				# update filter
				if newspan:
					if not cell[lhs]:
						del cell[lhs]
						continue
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right

			# unary rules
			unaryagenda.update(cell)
			while unaryagenda:
				rhs1 = unaryagenda.pop()
				for n in range(grammar.numrules):
					rule = &(grammar.unary[rhs1][n])
					if rule.rhs1 != rhs1:
						break
					lhs = rule.lhs
					edge = new_CFGEdge(0.0, rule, right)
					if lhs in cell:
						cell[lhs][edge] = edge
						continue
					unaryagenda.add(lhs)
					cell[lhs] = {edge: edge}
					# update filter
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right
			nonempty = list(filter(None, cell.values()))
			if nonempty:
				spans += 1
				items += len(nonempty)
				edges += sum(map(len, nonempty))
	msg = "chart spans %d, items %d, edges %d" % (spans, items, edges)
	if not chart[0][lensent].get(start):
		return chart, NONE, "no parse " + msg
	return chart, new_CFGChartItem(start, 0, lensent), msg


cdef populatepos(Grammar grammar, list chart, sent, tags,
		short [:, :] minleft, short [:, :] maxleft,
		short [:, :] minright, short [:, :] maxright):
	""" Assign all possible POS tags for a word, and apply all possible
	unary rules to them.

	:returns: a dictionary with the best scoring edges for each lhs,
		or None if no POS tag was found for this word. """
	cdef:
		EdgeAgenda unaryagenda = EdgeAgenda()
		Rule *rule
		LexicalRule lexrule
		UInt n, lhs, rhs1
		dict viterbicell
		short left, right
	viterbi = [{} for _ in sent]
	for left, word in enumerate(sent):  # assign POS tags
		tag = tags[left].encode('ascii') if tags else None
		right = left + 1
		cell = chart[left][right]
		viterbicell = viterbi[left]
		recognized = False
		for lexrule in grammar.lexicalbyword.get(word, ()):
			if lexrule.lhs not in cell:
				continue
			lhs = lexrule.lhs
			# if we are given gold tags, make sure we only allow matching
			# tags - after removing addresses introduced by the DOP reduction
			if (tag is None or grammar.tolabel[lhs] == tag
					or grammar.tolabel[lhs].startswith(tag + b'@')):
				edge = new_CFGEdge(lexrule.prob, NULL, right)
				viterbicell[lhs] = edge
				cell[lhs] = {edge: edge}
				recognized = True
				# update filter
				if left > minleft[lhs, right]:
					minleft[lhs, right] = left
				if left < maxleft[lhs, right]:
					maxleft[lhs, right] = left
				if right < minright[lhs, left]:
					minright[lhs, left] = right
				if right > maxright[lhs, left]:
					maxright[lhs, left] = right
		# NB: don't allow blocking of gold tags if given
		if not recognized:
			if tag is None or tag not in grammar.toid:
				return None, "not covered: %r" % (tag or word, )
			lhs = grammar.toid[tag]
			edge = new_CFGEdge(0.0, NULL, right)
			viterbicell[lhs] = edge
			cell[lhs] = {edge: edge}
			# update filter
			if left > minleft[lhs, right]:
				minleft[lhs, right] = left
			if left < maxleft[lhs, right]:
				maxleft[lhs, right] = left
			if right < minright[lhs, left]:
				minright[lhs, left] = right
			if right > maxright[lhs, left]:
				maxright[lhs, left] = right

		# unary rules on the span of this POS tag
		# NB: for this agenda, only the probabilities of the edges matter
		unaryagenda.update(viterbicell.items())
		while unaryagenda.length:
			rhs1 = unaryagenda.popentry().key
			for n in range(grammar.numrules):
				rule = &(grammar.unary[rhs1][n])
				if rule.rhs1 != rhs1:
					break
				elif rule.lhs not in cell:
					continue
				lhs = rule.lhs
				prob = rule.prob + (<CFGEdge>viterbicell[rhs1]).inside
				edge = new_CFGEdge(prob, rule, right)
				if cell[lhs]:
					if prob < (<CFGEdge>viterbicell[lhs]).inside:
						unaryagenda.setifbetter(lhs, <CFGEdge>edge)
						viterbicell[lhs] = edge
						cell[lhs][edge] = edge
					elif (edge not in cell[lhs] or
							prob < (<CFGEdge>cell[lhs][edge]).inside):
						cell[lhs][edge] = edge
					continue
				unaryagenda.setifbetter(lhs, <CFGEdge>edge)
				viterbicell[lhs] = edge
				cell[lhs] = {edge: edge}
				# update filter
				if left > minleft[lhs, right]:
					minleft[lhs, right] = left
				if left < maxleft[lhs, right]:
					maxleft[lhs, right] = left
				if right < minright[lhs, left]:
					minright[lhs, left] = right
				if right > maxright[lhs, left]:
					maxright[lhs, left] = right
	return viterbi, ''


def doinsideoutside(sent, Grammar grammar, inside=None, outside=None,
		tags=None, start=None):
	assert grammar.maxfanout == 1, "Not a PCFG! fanout = %d" % grammar.maxfanout
	assert not grammar.logprob, "Grammar must not have log probabilities."
	lensent = len(sent)
	if start is None:
		start = grammar.toid[grammar.start]
	if inside is None:
		inside = np.zeros((lensent, lensent + 1,
				grammar.nonterminals), dtype='d')
	else:
		inside[:len(sent), :len(sent) + 1, :] = 0.0
	if outside is None:
		outside = np.zeros((lensent, lensent + 1,
				grammar.nonterminals), dtype='d')
	else:
		outside[:len(sent), :len(sent) + 1, :] = 0.0
	minmaxlr = insidescores(sent, grammar, inside, tags)
	if inside[0, len(sent), start]:
		outsidescores(grammar, sent, start, inside, outside, *minmaxlr)
		msg = 'inside prob=%g' % inside[0, len(sent), start]
		start = new_CFGChartItem(start, 0, lensent)
	else:
		start = NONE
		msg = "no parse"
	return inside, outside, start, msg


def insidescores(sent, Grammar grammar, double [:, :, :] inside, tags=None):
	""" Compute inside scores. These are not viterbi scores, but sums of
	all derivations headed by a certain nonterminal + span. """
	cdef:
		short left, right, mid, span, lensent = len(sent)
		short narrowl, narrowr, minmid, maxmid
		double oldscore, prob, ls, rs, ins
		UInt n, lhs, rhs1
		bint foundbetter = False
		Rule *rule
		LexicalRule lexrule
		unicode word
		list cell = [{} for _ in grammar.toid]
		EdgeAgenda unaryagenda = EdgeAgenda()
		short [:, :] minleft, maxleft, minright, maxright
		double [:] unaryscores = np.empty(grammar.nonterminals, dtype='d')
	minleft, maxleft, minright, maxright = minmaxmatrices(
			grammar.nonterminals, lensent)
	for left in range(lensent):  # assign POS tags
		tag = tags[left].encode('ascii') if tags else None
		right = left + 1
		for lexrule in grammar.lexicalbyword.get(sent[left], []):
			lhs = lexrule.lhs
			# if we are given gold tags, make sure we only allow matching
			# tags - after removing addresses introduced by the DOP reduction
			if not tags or (grammar.tolabel[lhs] == tag
					or grammar.tolabel[lhs].startswith(tag + b'@')):
				inside[left, right, lhs] = lexrule.prob
		if not inside.base[left, right].any():
			if tags is not None:
				lhs = grammar.toid[tag]
				if not tags or (grammar.tolabel[lhs] == tag
						or grammar.tolabel[lhs].startswith(tag + b'@')):
					inside[left, right, lhs] = 1.
			else:
				raise ValueError("not covered: %r" % (tag or sent[left]), )
		# unary rules on POS tags (NB: agenda is a min-heap, negate probs)
		unaryagenda.update([(rhs1,
			new_CFGEdge(-inside[left, right, rhs1], NULL, 0))
			for rhs1 in range(grammar.nonterminals)
			if inside[left, right, rhs1]
			and grammar.unary[rhs1].rhs1 == rhs1])
		unaryscores[:] = 0.0
		while unaryagenda.length:
			rhs1 = unaryagenda.popentry().key
			for n in range(grammar.numrules):
				rule = &(grammar.unary[rhs1][n])
				if rule.rhs1 != rhs1:
					break
				prob = rule.prob * inside[left, right, rhs1]
				lhs = rule.lhs
				edge = new_CFGEdge(-prob, rule, right)
				if edge not in cell[lhs]:
					unaryagenda.setifbetter(lhs, <CFGEdge>edge)
					inside[left, right, lhs] += prob
					cell[lhs][edge] = edge
		for a in cell:
			a.clear()
		for lhs in range(grammar.nonterminals):
			if inside[left, right, lhs]:
				# update filter
				if left > minleft[lhs, right]:
					minleft[lhs, right] = left
				if left < maxleft[lhs, right]:
					maxleft[lhs, right] = left
				if right < minright[lhs, left]:
					minright[lhs, left] = right
				if right > maxright[lhs, left]:
					maxright[lhs, left] = right
	for span in range(2, lensent + 1):
		# constituents from left to right
		for left in range(lensent - span + 1):
			right = left + span
			# binary rules
			for n in range(grammar.numrules):
				rule = &(grammar.bylhs[0][n])
				lhs = rule.lhs
				if lhs == grammar.nonterminals:
					break
				elif not rule.rhs2:
					continue
				narrowr = minright[rule.rhs1, left]
				narrowl = minleft[rule.rhs2, right]
				if narrowr >= right or narrowl < narrowr:
					continue
				widel = maxleft[rule.rhs2, right]
				minmid = narrowr if narrowr > widel else widel
				wider = maxright[rule.rhs1, left]
				maxmid = wider if wider < narrowl else narrowl
				#oldscore = inside[left, right, lhs]
				foundbetter = False
				for split in range(minmid, maxmid + 1):
					ls = inside[left, split, rule.rhs1]
					if ls == 0.0:
						continue
					rs = inside[split, right, rule.rhs2]
					if rs == 0.0:
						continue
					foundbetter = True
					inside[left, right, lhs] += rule.prob * ls * rs
					#assert 0.0 < inside[left, right, lhs] <= 1.0, (
					#	inside[left, right, lhs],
					#	left, right, grammar.tolabel[lhs])
				if foundbetter:  # and oldscore == 0.0:
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right
			# unary rules on this span
			unaryagenda.update([(rhs1,
				new_CFGEdge(-inside[left, right, rhs1], NULL, 0))
				for rhs1 in range(grammar.nonterminals)
				if inside[left, right, rhs1]
				and grammar.unary[rhs1].rhs1 == rhs1])
			unaryscores[:] = 0.0
			while unaryagenda.length:
				rhs1 = unaryagenda.popentry().key
				for n in range(grammar.numrules):
					rule = &(grammar.unary[rhs1][n])
					if rule.rhs1 != rhs1:
						break
					prob = rule.prob * inside[left, right, rhs1]
					lhs = rule.lhs
					edge = new_CFGEdge(-prob, rule, right)
					if edge not in cell[lhs]:
						unaryagenda.setifbetter(lhs, <CFGEdge>edge)
						inside[left, right, lhs] += prob
						cell[lhs][edge] = edge
			for a in cell:
				a.clear()
			for lhs in range(grammar.nonterminals):
				# update filter
				if inside[left, right, lhs]:
					if left > minleft[lhs, right]:
						minleft[lhs, right] = left
					if left < maxleft[lhs, right]:
						maxleft[lhs, right] = left
					if right < minright[lhs, left]:
						minright[lhs, left] = right
					if right > maxright[lhs, left]:
						maxright[lhs, left] = right
	return minleft, maxleft, minright, maxright


def outsidescores(Grammar grammar, sent, UInt start,
		double [:, :, :] inside, double [:, :, :] outside,
		short [:, :] minleft, short [:, :] maxleft,
		short [:, :] minright, short [:, :] maxright):
	cdef:
		short left, right, mid, span, lensent = len(sent)
		short narrowl, narrowr, minmid, maxmid
		double ls, rs, os
		UInt n, lhs, rhs1, rhs2
		bint foundbetter = False
		Rule *rule
		LexicalRule lexrule
		EdgeAgenda unaryagenda = EdgeAgenda()
		list cell = [{} for _ in grammar.toid]
	outside[0, lensent, start] = 1.0
	for span in range(lensent, 0, -1):
		for left in range(1 + lensent - span):
			right = left + span
			# unary rules
			unaryagenda.update([(lhs,
				new_CFGEdge(-outside[left, right, lhs], NULL, 0))
				for lhs in range(grammar.nonterminals)
				if outside[left, right, lhs]])
			while unaryagenda.length:
				lhs = unaryagenda.popentry().key
				for n in range(grammar.numrules):
					rule = &(grammar.bylhs[lhs][n])
					if rule.lhs != lhs:
						break
					elif rule.rhs2:
						continue
					prob = rule.prob * outside[left, right, lhs]
					edge = new_CFGEdge(-prob, rule, right)
					if edge not in cell[lhs]:
						unaryagenda.setifbetter(rule.rhs1, <CFGEdge>edge)
						cell[lhs][edge] = edge
						outside[left, right, rule.rhs1] += prob
						#assert 0.0 < outside[left, right, rule.rhs1] <= 1.0, (
						#		'illegal value: outside[%d, %d, %s] = %g' % (
						#			left, right, grammar.tolabel[rule.rhs1],
						#			outside[left, right, rule.rhs1]),
						#		rule.prob, outside[left, right, lhs],
						#		grammar.tolabel[rule.lhs])
			for rhs1 in range(grammar.nonterminals):
				cell[rhs1].clear()
			# binary rules
			for n in range(grammar.numrules):
				rule = &(grammar.bylhs[0][n])
				lhs = rule.lhs
				if lhs == grammar.nonterminals:
					break
				elif not rule.rhs2 or outside[left, right, lhs] == 0.0:
					continue
				os = outside[left, right, lhs]
				narrowr = minright[rule.rhs1, left]
				narrowl = minleft[rule.rhs2, right]
				if narrowr >= right or narrowl < narrowr:
					continue
				widel = maxleft[rule.rhs2, right]
				minmid = narrowr if narrowr > widel else widel
				wider = maxright[rule.rhs1, left]
				maxmid = wider if wider < narrowl else narrowl
				for split in range(minmid, maxmid + 1):
					ls = inside[left, split, rule.rhs1]
					if ls == 0.0:
						continue
					rs = inside[split, right, rule.rhs2]
					if rs == 0.0:
						continue
					outside[left, split, rule.rhs1] += rule.prob * rs * os
					outside[split, right, rule.rhs2] += rule.prob * ls * os
					#assert 0.0 < outside[left, split, rule.rhs1] <= 1.0, (
					#		'illegal value: outside[%d, %d, %s] = %g' % (
					#			left, split, grammar.tolabel[rule.rhs1],
					#			outside[left, split, rule.rhs1]),
					#		rule.prob, rs, os, grammar.tolabel[rule.lhs])
					#assert 0.0 < outside[split, right, rule.rhs2] <= 1.0, (
					#		'illegal value: outside[%d, %d, %s] = %g' % (
					#			split, right, grammar.tolabel[rule.rhs2],
					#			outside[split, right, rule.rhs2]))


def minmaxmatrices(nonterminals, lensent):
	""" Create matrices for the filter which tracks minima and maxima for
	splits of binary rules. """
	minleft = np.empty((nonterminals, lensent + 1), dtype='int16')
	maxleft = np.empty_like(minleft)
	minleft[...], maxleft[...] = -1, lensent + 1
	minright, maxright = maxleft.copy(), minleft.copy()
	return minleft, maxleft, minright, maxright


def chartmatrix(nonterminals, lensent):
	viterbi = np.empty((nonterminals, lensent, lensent + 1), dtype='d')
	viterbi[...] = np.inf
	return viterbi


def pprint_chart(chart, sent, tolabel):
	cdef CFGEdge edge
	print("chart:")
	for left, _ in enumerate(chart):
		for right, _ in enumerate(chart[left]):
			for label in chart[left][right] or ():
				#if not chart[left][right][label]:
				#	continue
				print("%s[%d:%d] =>" % (
						tolabel[label].decode('ascii'), left, right))
				for edge in sorted(chart[left][right][label] or (),
						key=sortfunc):
					if edge.rule is NULL:
						print("%9.7f  %9.7f  '%s'" % (exp(-edge.inside), 1,
								sent[edge.mid - 1]), end='')
					else:
						print("%9.7f  %9.7f " % (exp(-edge.inside),
								exp(-edge.rule.prob)), "%s[%d:%d]" % (
									tolabel[edge.rule.rhs1].decode('ascii'),
								left, edge.mid), end='')
						if edge.rule.rhs2:
							print("\t%s[%d:%d]" % (
									tolabel[edge.rule.rhs2].decode('ascii'),
									edge.mid, right), end='')
					print()
				print()


def sortfunc(CFGEdge e):
	return e.inside


def pprint_matrix(matrix, sent, tolabel, matrix2=None):
	""" Print a numpy matrix chart; optionally in parallel with another. """
	for span in range(1, len(sent) + 1):
		for left in range(len(sent) - span + 1):
			right = left + span
			if matrix[left, right].any() or (
					matrix2 is not None and matrix2[left, right].any()):
				print("[%d:%d]" % (left, right))
				for lhs in range(len(matrix[left, right])):
					if matrix[left, right, lhs] or (
							matrix2 is not None and matrix2[left, right, lhs]):
						print("%20s\t%8.6g" % (tolabel[lhs].decode('ascii'),
								matrix[left, right, lhs]), end='')
						if matrix2 is not None:
							print("\t%8.6g" % matrix2[left, right, lhs], end='')
						print()


def main():
	from discodop._grammar import Grammar
	from discodop.disambiguation import marginalize
	from operator import itemgetter
	cdef Rule rule
	cfg = Grammar([
		((('A', 'A'), ((0, ), )), 0.7), ((('A', 'B'), ((0, ), )), 0.6),
		((('A', 'C'), ((0, ), )), 0.5), ((('A', 'D'), ((0, ), )), 0.4),
		((('B', 'A'), ((0, ), )), 0.3), ((('B', 'B'), ((0, ), )), 0.2),
		((('B', 'C'), ((0, ), )), 0.1), ((('B', 'D'), ((0, ), )), 0.2),
		((('B', 'C'), ((0, ), )), 0.3), ((('C', 'A'), ((0, ), )), 0.4),
		((('C', 'B'), ((0, ), )), 0.5), ((('C', 'C'), ((0, ), )), 0.6),
		((('C', 'D'), ((0, ), )), 0.7), ((('D', 'A'), ((0, ), )), 0.8),
		((('D', 'B'), ((0, ), )), 0.9), ((('D', 'NP', 'VP'), ((0, 1), )), 1),
		((('D', 'C'), ((0, ), )), 0.8), ((('S', 'D'), ((0, ), )), 0.5),
		((('S', 'A'), ((0, ), )), 0.8), ((('NP', 'Epsilon'), ('mary', )), 1),
		((('VP', 'Epsilon'), ('walks', )), 1)],
		start='S')
	print(cfg)
	print("cfg parsing; sentence: mary walks")
	print("pcfg", end='')
	chart, start, msg = parse_sparse("mary walks".split(), cfg)
	assert start, msg
	chart, start, msg = parse_dense("mary walks".split(), cfg)
	assert start, msg
	pprint_chart(chart, "mary walks".split(), cfg.tolabel)
	cfg1 = Grammar([
		((('NP', 'Epsilon'), ('mary', )), 1),
		((('S', 'NP', 'VP'), ((0, 1), )), 1),
		((('VP', 'Epsilon'), ('walks', )), 1)], start='S')
	cfg1.switch('default', False)
	i, o, start, _ = doinsideoutside("mary walks".split(), cfg1)
	assert start
	print(i[0, 2, cfg1.toid[b'S']], o[0, 2, cfg1.toid[b'S']])
	i, o, start, _ = doinsideoutside("walks mary".split(), cfg1)
	assert not start
	print(i[0, 2, cfg1.toid[b'S']], o[0, 2, cfg1.toid[b'S']])
	rules = [
		((('NP', 'NP', 'PP'), ((0, 1), )), 0.4),
		((('PP', 'P', 'NP'), ((0, 1), )), 1),
		((('S', 'NP', 'VP'), ((0, 1), )), 1),
		((('VP', 'V', 'NP'), ((0, 1), )), 0.7),
		((('VP', 'VP', 'PP'), ((0, 1), )), 0.3),
		((('NP', 'Epsilon'), ('astronomers', )), 0.1),
		((('NP', 'Epsilon'), ('ears', )), 0.18),
		((('V', 'Epsilon'), ('saw', )), 1),
		((('NP', 'Epsilon'), ('saw', )), 0.04),
		((('NP', 'Epsilon'), ('stars', )), 0.18),
		((('NP', 'Epsilon'), ('telescopes', )), 0.1),
		((('P', 'Epsilon'), ('with', )), 1)]
	cfg2 = Grammar(rules, start='S')
	cfg2.switch('default', False)
	sent = "astronomers saw stars with ears".split()
	inside, outside, _, msg = doinsideoutside(sent, cfg2)
	print(msg)
	pprint_matrix(inside, sent, cfg2.tolabel, outside)
	cfg2.switch('default', True)
	chart, start, msg = parse(sent, cfg2)
	mpp, _, _ = marginalize('mpp', chart, start, cfg2, 10)
	for a, p in sorted(mpp.items(), key=itemgetter(1), reverse=True):
		print(p, a)
	chart1, start1, msg1 = parse_symbolic(sent, cfg2)
	print(msg, '\n', msg1)
